import pickle
from copy import deepcopy
from collections import OrderedDict
from function import *
from function_class import *
from gpu import *
from utils import *


class BaseModel:
    def __init__(self):
        """Base object for Model
        """
        self.network = OrderedDict()

    def __call__(self, arg):
        result = self.forward(arg)
        return result

    def forward():
        pass


class BaseLayer():
    def __init__(self):
        """Base object for Layer
        """
        self.differentiable = False
        self.changeability = False
        self.mixed_precision = False
        self.tying = False
        self.tied = False
        self.parameter = OrderedDict()

    def _to_cpu(self):
        """Layer의 parameter를 vram에서 ram으로 옮김
        """
        param_list = list(self.parameter.keys())
        for param in param_list:
            self.parameter[param] = to_cpu(self.parameter[param])

    def _to_gpu(self):
        """Layer의 parameter를 ram에서 vram으로 옮김
        """
        param_list = list(self.parameter.keys())
        for param in param_list:
            self.parameter[param] = to_gpu(self.parameter[param])

    def _save_params(self):
        """Layer의 parameter 반환

        Returns:
            OrderedDict
        """
        return deepcopy(self.parameter)

    def _load_params(self, params):
        self.parameter = deepcopy(params)

    def _mixed_precision_training(self):
        self.mixed_precision = True

    def _mixed_precision_training_off(self):
        self.mixed_precision = False


class DenseLayer(BaseLayer):
    def __init__(self, input_size: int, output_size: int, initialize: str = "He"):
        """Initialize DenseLayer

        Parameters
        ----------
        input_size (int) : input node 개수

        output_size (int) : output node 개수

        initialize (str, optional) : 가중치 초기화 방법 설정. Default: "He"

        """
        super().__init__()
        self.input_size = input_size
        self.output_size = output_size
        self.differentiable = True
        self.x = None
        # 4-dim tensor 대응
        self.orgin_x_shape = None

        if initialize == "He":
            self.parameter["weight"] = np.random.randn(input_size, output_size).astype(np.float32) * (
                np.sqrt(2 / input_size))
            self.parameter["bias"] = np.zeros(output_size).astype(np.float32)

        elif initialize == "Xavier":
            self.parameter["weight"] = np.random.randn(input_size, output_size).astype(np.float32) * np.sqrt(
                1 / input_size)
            self.parameter["bias"] = np.zeros(output_size).astype(np.float32)

        elif initialize == "None":
            self.parameter["weight"] = 0.01 * np.random.randn(input_size, output_size).astype(np.float32)
            self.parameter["bias"] = np.zeros(output_size).astype(np.float32)

        elif isinstance(initialize, int):
            self.parameter["weight"] = np.random.randn(input_size, output_size).astype(np.float32) * np.sqrt(
                1 / initialize)
            self.parameter["bias"] = np.zeros(output_size).astype(np.float32)

        else:
            raise ValueError("'initialize' must be 'He' or 'Xavier' or 'None' or integer")

        self.dw = np.zeros_like(self.parameter["weight"])
        self.db = np.zeros_like(self.parameter["bias"])

    def __call__(self, arg: np.ndarray):
        """Wrapper function for _forward()

        Parameters
        ----------
        Args:
            arg (numpy.ndarray): 이전 layer의 forward 결과

        Returns:
            numpy.ndarray
        """
        result = self._forward(arg)
        return result

    def __repr__(self):
        return "DenseLayer"

    def _forward(self, x: np.ndarray) -> np.ndarray:
        """DenseLayer의 feedforward

        Args:
            x (np.ndarray): 이전 layer의 feedforward 결과

        Returns:
            np.ndarray: DenseLayer의 feedforward 결과
        """
        if self.mixed_precision:
            # sequnce 대응
            if x.ndim == 3:
                batch_size, n_timestep, _ = x.shape
                reshaped_x = x.reshape(batch_size * n_timestep, -1)
                self.x = x
                result = np.matmul(reshaped_x, self.parameter["weight"].astype(np.float16)) + self.parameter[
                    "bias"].astype(np.float16)
                result = result.reshape(batch_size, n_timestep, -1)

                return result

            # 4-dim tensor 대응
            self.origin_x_shape = x.shape
            x = x.reshape(x.shape[0], -1)
            self.x = x
            result = np.matmul(x, self.parameter["weight"].astype(np.float16)) + self.parameter["bias"].astype(
                np.float16)

        else:
            # sequnce 대응
            if x.ndim == 3:
                batch_size, n_timestep, _ = x.shape
                reshaped_x = x.reshape(batch_size * n_timestep, -1)
                self.x = x
                result = np.matmul(reshaped_x, self.parameter["weight"]) + self.parameter["bias"]
                result = result.reshape(batch_size, n_timestep, -1)

                return result

            # 4-dim tensor 대응
            self.origin_x_shape = x.shape
            x = x.reshape(x.shape[0], -1)
            self.x = x
            result = np.matmul(x, self.parameter["weight"]) + self.parameter["bias"]

        return result

    def _backward(self, input: np.ndarray) -> np.ndarray:
        """DenseLayer의 backpropagation

        Args:
            input (np.ndarray): 이전 layer의 backpropagation 결과

        Returns:
            np.ndarray: DenseLayer의 backpropagation 결과
        """
        if self.mixed_precision:
            if self.x.ndim == 3:
                x = self.x
                batch_size, n_timestep, _ = x.shape
                input = input.reshape(batch_size * n_timestep, -1)
                reshaped_x = x.reshape(batch_size * n_timestep, -1)

                db = np.sum(input, axis=0)
                dw = np.matmul(reshaped_x.T, input)
                dx = np.matmul(input, self.parameter["weight"].T.astype(np.float16))
                dx = dx.reshape(*x.shape)

                self.dw[...] = dw
                self.db[...] = db

                return dx

            else:
                dx = np.matmul(input, self.parameter["weight"].T.astype(np.float16))
                self.dw = np.matmul(self.x.T, input)
                self.db = np.sum(input, axis=0)

                # for 4-dim tensor
                dx = dx.reshape(*self.origin_x_shape)

        else:
            if self.x.ndim == 3:
                x = self.x
                batch_size, n_timestep, _ = x.shape
                input = input.reshape(batch_size * n_timestep, -1)
                reshaped_x = x.reshape(batch_size * n_timestep, -1)

                db = np.sum(input, axis=0)
                dw = np.matmul(reshaped_x.T, input)
                dx = np.matmul(input, self.parameter["weight"].T)
                dx = dx.reshape(*x.shape)

                self.dw[...] = dw
                self.db[...] = db

                return dx

            else:
                dx = np.matmul(input, self.parameter["weight"].T)
                self.dw = np.matmul(self.x.T, input)
                self.db = np.sum(input, axis=0)

                # for 4-dim tensor
                dx = dx.reshape(*self.origin_x_shape)

        return dx

    def load_parameter(self, parameter: tuple):
        weight, bias = parameter
        self.parameter["weight"] = np.array(weight)
        self.parameter["bias"] = np.array(bias)

    def get_gradient(self) -> OrderedDict:
        """Layer의 gradient를 return

        Returns:
            OrderedDict: Layer의 gradient
        """
        grad = OrderedDict()
        grad["dw"] = self.dw.astype(np.float32)
        grad["db"] = self.db.astype(np.float32)

        return grad


class Embedding(BaseLayer):
    def __init__(self, parameter, mixed_precision=False):
        """
        Initialize EmbeddingLayer

        Parameters
        ----------
        vocab_size (int) : vocab의 size

        hidden_size (int) : embedding될 representation의 hidden size

        initialize (str, optional) : 가중치 초기화 방법 설정. Default: "He"

        """
        super().__init__()
        self.differentiable = True
        self.index = None
        self.parameter["weight"] = parameter["weight"]
        self.dw = np.zeros_like(self.parameter["weight"])
        self.mixed_precision = mixed_precision

    def __call__(self, arg):
        result = self._forward(arg)
        return result

    def __repr__(self):
        return "Embedding"

    def _forward(self, index):
        if self.mixed_precision:
            self.index = index
            weight = self.parameter["weight"].astype(np.float16)
            result = weight[index]

        else:
            self.index = index
            weight = self.parameter["weight"]
            result = weight[index]

        return result

    def _backward(self, input):
        """

        Parameter
        ---------
        input : 다음 layer의 해당 layer의 output에 대한 미분값

        Variable
        --------
        dx = forward시 input에 대한 미분값으로 이전 layer로 넘겨준다.
        self.dw = weight에 대한 미분값
        self.db = bias에 대한 미분값

        """
        if self.mixed_precision:
            dw = self.dw
            dw[...] = 0
            if GPU:
                import cupyx
                cupyx.scatter_add(dw, self.index, input)
            else:
                np.add.at(dw, self.index, input)

        else:
            dw = self.dw
            dw[...] = 0
            if GPU:
                import cupyx
                cupyx.scatter_add(dw, self.index, input)
            else:
                np.add.at(dw, self.index, input)

        return None

    def get_gradient(self) -> OrderedDict:
        """Layer의 gradient를 return

        Returns:
            OrderedDict: Layer의 gradient
        """
        grad = OrderedDict()
        grad["dw"] = self.dw.astype(np.float32)

        return grad

    def _get_gradient(self):
        dw = self.dw.astype(np.float32)

        return dw


class EmbeddingDot(BaseLayer):
    def __init__(self, vocab_size, hidden_size, initialize="He"):
        super().__init__()
        self.differentiable = True
        self.cache = None
        self.mixed_precision = False

        if initialize == "He":
            self.parameter["weight"] = np.random.randn(vocab_size, hidden_size).astype(np.float32) * (
                np.sqrt(2 / vocab_size))

        elif initialize == "Xavier":
            self.parameter["weight"] = np.random.randn(vocab_size, hidden_size).astype(np.float32) * np.sqrt(
                1 / vocab_size)

        elif initialize == "None":
            self.parameter["weight"] = 0.01 * np.random.randn(vocab_size, hidden_size).astype(np.float32)

        elif isinstance(initialize, int):
            self.parameter["weight"] = (1 / initialize) * np.random.randn(vocab_size, hidden_size).astype(np.float32)

        else:
            raise ValueError("'initialize' must be 'He' or 'Xavier' or 'None' or integer")

        self.embed = Embedding(self.parameter, mixed_precision=False)

        self.dw = None

    def __repr__(self) -> str:
        return "EmbeddingDot"

    def __call__(self, arg):
        result = self._forward(arg)
        return result

    def _forward(self, h, index):
        target_weight = self.embed._forward(index)
        out = np.sum(target_weight * h, axis=1)
        self.cache = (h, target_weight)

        return out

    def _backward(self, dout):
        h, target_weight = self.cache
        dout = dout.reshape(dout.shape[0], 1)
        dtarget_W = dout * h
        self.embed._backward(dtarget_W)
        self.dw = self.embed._get_gradient()
        dh = dout * target_weight

        return dh

    def get_gradient(self) -> OrderedDict:
        """Layer의 gradient를 return

        Returns:
            OrderedDict: Layer의 gradient
        """
        grad = OrderedDict()
        grad["dw"] = self.dw.astype(np.float32)

        return grad


class EmbeddingLayer(BaseLayer):
    def __init__(self, vocab_size: int, hidden_size: int, initialize="He"):
        super().__init__()
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.differentiable = True

        if initialize == "He":
            self.parameter["weight"] = np.random.randn(vocab_size, hidden_size).astype(np.float32) * (
                np.sqrt(2 / vocab_size))

        elif initialize == "Xavier":
            self.parameter["weight"] = np.random.randn(vocab_size, hidden_size).astype(np.float32) * np.sqrt(
                1 / vocab_size)

        elif initialize == "None":
            self.parameter["weight"] = 0.01 * np.random.randn(vocab_size, hidden_size).astype(np.float32)

        elif isinstance(initialize, int):
            self.parameter["weight"] = (1 / initialize) * np.random.randn(vocab_size, hidden_size).astype(np.float32)

        else:
            raise ValueError("'initialize' must be 'He' or 'Xavier' or 'None' or integer")

        self.layer = None
        self.dw = np.zeros_like(self.parameter["weight"]).astype(np.float32)

    def __call__(self, arg):
        result = self._forward(arg)
        return result

    def __repr__(self):
        return "EmbeddingLayer"

    def _forward(self, x):
        # Mixed Precision Training
        if self.mixed_precision:
            batch_size, n_timestep = x.shape

            result = np.empty((batch_size, n_timestep, self.hidden_size), dtype="f")
            self.layer = []

            for timestep in range(n_timestep):
                embedding_cell = Embedding(self.parameter, mixed_precision=True)
                result[:, timestep, :] = embedding_cell._forward(x[:, timestep])
                self.layer.append(embedding_cell)

        else:
            batch_size, n_timestep = x.shape

            result = np.empty((batch_size, n_timestep, self.hidden_size), dtype="f")
            self.layer = []

            for timestep in range(n_timestep):
                embedding_cell = Embedding(self.parameter)
                result[:, timestep, :] = embedding_cell._forward(x[:, timestep])
                self.layer.append(embedding_cell)

        return result

    def _backward(self, dout):
        # Mixed Precision Training
        if self.mixed_precision:
            w = self.parameter["weight"].astype(np.float16)
            batch_size, n_timestep, hidden_size = dout.shape

            dw = np.zeros_like(w).astype(np.float32)
            for timestep in range(n_timestep):
                embedding_cell = self.layer[timestep]
                embedding_cell._backward(dout[:, timestep, :])
                temp_dw = embedding_cell._get_gradient()
                dw += temp_dw

            self.dw[...] = dw

        else:
            w = self.parameter["weight"]
            batch_size, n_timestep, hidden_size = dout.shape

            dw = np.zeros_like(w).astype(np.float32)
            for timestep in range(n_timestep):
                embedding_cell = self.layer[timestep]
                embedding_cell._backward(dout[:, timestep, :])
                temp_dw = embedding_cell._get_gradient()
                dw += temp_dw

            self.dw[...] = dw

        return None

    def get_gradient(self) -> OrderedDict:
        """Layer의 gradient를 return

        Returns:
            OrderedDict: Layer의 gradient
        """
        grad = OrderedDict()
        grad["dw"] = self.dw.astype(np.float32)

        return grad

    def _fp16_grad(self):
        self.dw = np.zeros_like(self.parameter["weight"]).astype(np.float16)


class ConvLayer(BaseLayer):
    """
    Convolution layer
    """

    def __init__(self, input_channel: int, output_channel: int, kernel_size: int, stride=1, padding=0, initialize="He"):
        """
        Initialize ConvLayer

        Parameters
        ----------
        input_channel (int) : input의 channel 개수

        output_channel (int) : kernel의 개수

        kernel_size (int or tuple) : kernel(height, width)의 크기

        stride (int, optional) : stride 설정. Default: 1

        padding (int, optional) : padding 설정. Default: 0

        initialize (str, optional) : 가중치 초기화 방법 설정. Default: "He"


        """
        super().__init__()
        self.differentiable = True
        self.input_channel = input_channel
        self.output_channel = output_channel

        if isinstance(kernel_size, int):
            self.kernel_width = kernel_size
            self.kernel_height = kernel_size

        elif isinstance(kernel_size, tuple):
            self.kernel_height, self.kernel_width = kernel_size

        self.stride = stride
        self.padding = padding
        self.fan_in = self.kernel_width * self.kernel_height * input_channel
        self.fan_out = self.kernel_width * self.kernel_height * output_channel

        self.x = None
        self.col = None
        self.col_weight = None
        self.dw = None
        self.db = None

        if initialize == "He":
            self.parameter["weight"] = np.random.randn(self.output_channel, self.input_channel, self.kernel_height,
                                                       self.kernel_width).astype(np.float32) * (
                                           np.sqrt(2 / self.fan_in))
            self.parameter["bias"] = np.zeros(self.output_channel).astype(np.float32)

        elif initialize == "Xavier":
            self.parameter["weight"] = np.random.randn(self.output_channel, self.input_channel, self.kernel_height,
                                                       self.kernel_width).astype(np.float32) * np.sqrt(1 / self.fan_in)
            self.parameter["bias"] = np.zeros(self.output_channel).astype(np.float32)

        elif initialize == "None":
            self.parameter["weight"] = 0.01 * np.random.randn(self.output_channel, self.input_channel,
                                                              self.kernel_height, self.kernel_width).astype(np.float32)
            self.parameter["bias"] = np.zeros(self.output_channel).astype(np.float32)

        else:
            raise ValueError("'initialize' must be 'He' or 'Xavier' or 'None'")

    def __repr__(self):
        return "ConvLayer"

    def __call__(self, arg):
        result = self._forward(arg)
        return result

    def _forward(self, x):
        # Mixed Precision Training
        if self.mixed_precision:
            n_input, n_input_channel, input_height, input_width = x.shape
            out_height = int(1 + (input_height + self.padding * 2 - self.kernel_height) / self.stride)
            out_width = int(1 + (input_width + self.padding * 2 - self.kernel_width) / self.stride)

            col = self.img2col(x)
            col_weight = self.parameter["weight"].reshape(self.output_channel, -1).T.astype(np.float16)
            result = np.matmul(col, col_weight) + self.parameter["bias"].astype(np.float16)
            result = result.reshape(n_input, out_height, out_width, -1).transpose(0, 3, 1, 2)

            self.x = x
            self.col = col
            self.col_weight = col_weight

        else:
            n_input, n_input_channel, input_height, input_width = x.shape
            out_height = int(1 + (input_height + self.padding * 2 - self.kernel_height) / self.stride)
            out_width = int(1 + (input_width + self.padding * 2 - self.kernel_width) / self.stride)

            col = self.img2col(x)
            col_weight = self.parameter["weight"].reshape(self.output_channel, -1).T
            result = np.matmul(col, col_weight) + self.parameter["bias"]
            result = result.reshape(n_input, out_height, out_width, -1).transpose(0, 3, 1, 2)

            self.x = x
            self.col = col
            self.col_weight = col_weight

        return result

    def _backward(self, input):
        # Mixed Precision Training
        if self.mixed_precision:
            input = input.transpose(0, 2, 3, 1).reshape(-1, self.output_channel)

            self.dw = np.matmul(self.col.T, input)
            self.dw = self.dw.transpose(1, 0).reshape(self.output_channel, self.input_channel, self.kernel_height,
                                                      self.kernel_width)
            self.db = np.sum(input, axis=0)

            dcol = np.matmul(input, self.col_weight.T)
            result = self.col2img(dcol, self.x.shape)

        else:
            input = input.transpose(0, 2, 3, 1).reshape(-1, self.output_channel)

            self.dw = np.matmul(self.col.T, input)
            self.dw = self.dw.transpose(1, 0).reshape(self.output_channel, self.input_channel, self.kernel_height,
                                                      self.kernel_width)
            self.db = np.sum(input, axis=0)

            dcol = np.matmul(input, self.col_weight.T)
            result = self.col2img(dcol, self.x.shape)

        return result

    def get_gradient(self) -> OrderedDict:
        """Layer의 gradient를 return

        Returns:
            OrderedDict: Layer의 gradient
        """
        grad = OrderedDict()
        grad["dw"] = self.dw.astype(np.float32)
        grad["db"] = self.db.astype(np.float32)

        return grad

    def _fp16_grad(self):
        pass

    def img2col(self, input_data):
        # Mixed Precision Training
        if self.mixed_precision:
            n_input, n_input_channel, input_height, input_width = input_data.shape
            out_height = int((input_height + self.padding * 2 - self.kernel_height) // self.stride + 1)
            out_width = int((input_width + self.padding * 2 - self.kernel_width) // self.stride + 1)

            img = np.pad(input_data, [(0, 0), (0, 0), (self.padding, self.padding), (self.padding, self.padding)],
                         'constant')
            col = np.zeros(
                (n_input, n_input_channel, self.kernel_height, self.kernel_width, out_height, out_width)).astype(
                np.float32)

            for y in range(self.kernel_height):
                y_max = y + self.stride * out_height
                for x in range(self.kernel_width):
                    x_max = x + self.stride * out_width
                    col[:, :, y, x, :, :] = img[:, :, y:y_max:self.stride, x:x_max:self.stride]

            col = col.transpose(0, 4, 5, 1, 2, 3).reshape(n_input * out_height * out_width, -1)

        else:
            n_input, n_input_channel, input_height, input_width = input_data.shape
            out_height = int((input_height + self.padding * 2 - self.kernel_height) // self.stride + 1)
            out_width = int((input_width + self.padding * 2 - self.kernel_width) // self.stride + 1)

            img = np.pad(input_data, [(0, 0), (0, 0), (self.padding, self.padding), (self.padding, self.padding)],
                         'constant')
            col = np.zeros(
                (n_input, n_input_channel, self.kernel_height, self.kernel_width, out_height, out_width)).astype(
                np.float32)

            for y in range(self.kernel_height):
                y_max = y + self.stride * out_height
                for x in range(self.kernel_width):
                    x_max = x + self.stride * out_width
                    col[:, :, y, x, :, :] = img[:, :, y:y_max:self.stride, x:x_max:self.stride]

            col = col.transpose(0, 4, 5, 1, 2, 3).reshape(n_input * out_height * out_width, -1)

        return col

    def col2img(self, col, input_shape):
        # Mixed Precision Training
        if self.mixed_precision:
            n_input, n_input_channel, input_height, input_width = input_shape
            out_height = int((input_height + 2 * self.padding - self.kernel_height) // self.stride + 1)
            out_width = int((input_width + 2 * self.padding - self.kernel_width) // self.stride + 1)
            col = col.reshape(n_input, out_height, out_width, n_input_channel, self.kernel_height,
                              self.kernel_width).transpose(0, 3, 4, 5, 1, 2)

            img = np.zeros((n_input, n_input_channel, input_height + 2 * self.padding + self.stride - 1,
                            input_width + 2 * self.padding + self.stride - 1)).astype(np.float32)
            for y in range(self.kernel_height):
                y_max = y + self.stride * out_height
                for x in range(self.kernel_width):
                    x_max = x + self.stride * out_width
                    img[:, :, y:y_max:self.stride, x:x_max:self.stride] += col[:, :, y, x, :, :]

        else:
            n_input, n_input_channel, input_height, input_width = input_shape
            out_height = int((input_height + 2 * self.padding - self.kernel_height) // self.stride + 1)
            out_width = int((input_width + 2 * self.padding - self.kernel_width) // self.stride + 1)
            col = col.reshape(n_input, out_height, out_width, n_input_channel, self.kernel_height,
                              self.kernel_width).transpose(0, 3, 4, 5, 1, 2)

            img = np.zeros((n_input, n_input_channel, input_height + 2 * self.padding + self.stride - 1,
                            input_width + 2 * self.padding + self.stride - 1)).astype(np.float32)
            for y in range(self.kernel_height):
                y_max = y + self.stride * out_height
                for x in range(self.kernel_width):
                    x_max = x + self.stride * out_width
                    img[:, :, y:y_max:self.stride, x:x_max:self.stride] += col[:, :, y, x, :, :]

        return img[:, :, self.padding:input_height + self.padding, self.padding:input_width + self.padding]


class GlobalAveragePoolingLayer(BaseLayer):
    def __init__(self):
        super().__init__()
        self.x = None

    def __repr__(self):
        return "GlobalAveragePoolingLayer"

    def __call__(self, arg):
        result = self._forward(arg)
        return result

    def _forward(self, x):
        n_input, n_input_channel, input_height, input_width = x.shape
        self.x = x

        reshaped_x = x.reshape(n_input, n_input_channel, -1)

        # result = (n_input, n_input_channel)
        result = reshaped_x.mean(axis=2)

        return result

    def _backward(self, input):
        # input = (n_input, n_input_channel)
        n_input, n_input_channel, input_height, input_width = self.x.shape
        temp_result = input * (1 / input_height * input_width)

        temp_reshaped = temp_result.reshape(n_input, n_input_channel, 1, 1)
        temp_reshaped = np.repeat(temp_reshaped, input_width, axis=3)
        result = np.repeat(temp_reshaped, input_height, axis=2)

        return result


class MaxPoolingLayer(BaseLayer):
    def __init__(self, kernel_size, stride=1, padding=0):
        super().__init__()

        if isinstance(kernel_size, int):
            self.kernel_width = kernel_size
            self.kernel_height = kernel_size

        elif isinstance(kernel_size, tuple):
            self.kernel_height, self.kernel_width = kernel_size

        self.stride = stride
        self.padding = padding

        self.x = None
        self.mask = None

    def __repr__(self):
        return "MaxPoolingLayer"

    def __call__(self, arg):
        result = self._forward(arg)
        return result

    def _forward(self, x):
        # Mixed Precision Training
        if self.mixed_precision:
            n_input, n_input_channel, input_height, input_width = x.shape

            out_height = int(1 + (input_height + self.padding * 2 - self.kernel_height) / self.stride)
            out_width = int(1 + (input_width + self.padding * 2 - self.kernel_width) / self.stride)

            col = self.img2col(x)
            col = col.reshape(-1, self.kernel_height * self.kernel_width)

            self.x = x
            self.mask = np.argmax(col, axis=1)  # (-1, self.kernel_height*kernel_width)
            result = np.max(col, axis=1)
            result = result.reshape(n_input, out_height, out_width, n_input_channel).transpose(0, 3, 1, 2)

        else:
            n_input, n_input_channel, input_height, input_width = x.shape

            out_height = int(1 + (input_height + self.padding * 2 - self.kernel_height) / self.stride)
            out_width = int(1 + (input_width + self.padding * 2 - self.kernel_width) / self.stride)

            col = self.img2col(x)
            col = col.reshape(-1, self.kernel_height * self.kernel_width)

            self.x = x
            self.mask = np.argmax(col, axis=1)  # (-1, self.kernel_height*kernel_width)
            result = np.max(col, axis=1)
            result = result.reshape(n_input, out_height, out_width, n_input_channel).transpose(0, 3, 1, 2)

        return result

    def _backward(self, input):
        # Mixed Precision Training
        if self.mixed_precision:
            input = input.transpose(0, 2, 3,
                                    1)  # (n_input, n_input_channel, input_height, input_width) -> (n_input, out_height, out_width, n_input_channel)
            kernel_size = self.kernel_height * self.kernel_width
            dmax = np.zeros((input.size, kernel_size)).astype(
                np.float32)  # (n_input*n_input_channel*input_height*input_width, self.kernel_height*kernel_width)
            dmax[np.arange(self.mask.size), self.mask.flatten()] = input.flatten()
            dmax = dmax.reshape(input.shape + (kernel_size,))

            dcol = dmax.reshape(dmax.shape[0] * dmax.shape[1] * dmax.shape[2], -1)
            result = self.col2img(dcol, self.x.shape)

        else:
            input = input.transpose(0, 2, 3,
                                    1)  # (n_input, n_input_channel, input_height, input_width) -> (n_input, out_height, out_width, n_input_channel)
            kernel_size = self.kernel_height * self.kernel_width
            dmax = np.zeros((input.size, kernel_size)).astype(
                np.float32)  # (n_input*n_input_channel*input_height*input_width, self.kernel_height*kernel_width)
            dmax[np.arange(self.mask.size), self.mask.flatten()] = input.flatten()
            dmax = dmax.reshape(input.shape + (kernel_size,))

            dcol = dmax.reshape(dmax.shape[0] * dmax.shape[1] * dmax.shape[2], -1)
            result = self.col2img(dcol, self.x.shape)

        return result

    def img2col(self, input_data):
        # Mixed Precision Training
        if self.mixed_precision:
            n_input, n_input_channel, input_height, input_width = input_data.shape
            out_height = (input_height + self.padding * 2 - self.kernel_height) // self.stride + 1
            out_width = (input_width + self.padding * 2 - self.kernel_width) // self.stride + 1

            img = np.pad(input_data, [(0, 0), (0, 0), (self.padding, self.padding), (self.padding, self.padding)],
                         'constant')
            col = np.zeros(
                (n_input, n_input_channel, self.kernel_height, self.kernel_width, out_height, out_width)).astype(
                np.float32)

            for y in range(self.kernel_height):
                y_max = y + self.stride * out_height
                for x in range(self.kernel_width):
                    x_max = x + self.stride * out_width
                    col[:, :, y, x, :, :] = img[:, :, y:y_max:self.stride, x:x_max:self.stride]

            col = col.transpose(0, 4, 5, 1, 2, 3).reshape(n_input * out_height * out_width, -1)

        else:
            n_input, n_input_channel, input_height, input_width = input_data.shape
            out_height = (input_height + self.padding * 2 - self.kernel_height) // self.stride + 1
            out_width = (input_width + self.padding * 2 - self.kernel_width) // self.stride + 1

            img = np.pad(input_data, [(0, 0), (0, 0), (self.padding, self.padding), (self.padding, self.padding)],
                         'constant')
            col = np.zeros(
                (n_input, n_input_channel, self.kernel_height, self.kernel_width, out_height, out_width)).astype(
                np.float32)

            for y in range(self.kernel_height):
                y_max = y + self.stride * out_height
                for x in range(self.kernel_width):
                    x_max = x + self.stride * out_width
                    col[:, :, y, x, :, :] = img[:, :, y:y_max:self.stride, x:x_max:self.stride]

            col = col.transpose(0, 4, 5, 1, 2, 3).reshape(n_input * out_height * out_width, -1)

        return col

    def col2img(self, col, input_shape):
        # Mixed Precision Training
        if self.mixed_precision:
            n_input, n_input_channel, input_height, input_width = input_shape
            out_height = (input_height + 2 * self.padding - self.kernel_height) // self.stride + 1
            out_width = (input_width + 2 * self.padding - self.kernel_width) // self.stride + 1
            col = col.reshape(n_input, out_height, out_width, n_input_channel, self.kernel_height,
                              self.kernel_width).transpose(0, 3, 4, 5, 1, 2)

            img = np.zeros((n_input, n_input_channel, input_height + 2 * self.padding + self.stride - 1,
                            input_width + 2 * self.padding + self.stride - 1)).astype(np.float32)
            for y in range(self.kernel_height):
                y_max = y + self.stride * out_height
                for x in range(self.kernel_width):
                    x_max = x + self.stride * out_width
                    img[:, :, y:y_max:self.stride, x:x_max:self.stride] += col[:, :, y, x, :, :]

        else:
            n_input, n_input_channel, input_height, input_width = input_shape
            out_height = (input_height + 2 * self.padding - self.kernel_height) // self.stride + 1
            out_width = (input_width + 2 * self.padding - self.kernel_width) // self.stride + 1
            col = col.reshape(n_input, out_height, out_width, n_input_channel, self.kernel_height,
                              self.kernel_width).transpose(0, 3, 4, 5, 1, 2)

            img = np.zeros((n_input, n_input_channel, input_height + 2 * self.padding + self.stride - 1,
                            input_width + 2 * self.padding + self.stride - 1)).astype(np.float32)
            for y in range(self.kernel_height):
                y_max = y + self.stride * out_height
                for x in range(self.kernel_width):
                    x_max = x + self.stride * out_width
                    img[:, :, y:y_max:self.stride, x:x_max:self.stride] += col[:, :, y, x, :, :]

        return img[:, :, self.padding:input_height + self.padding, self.padding:input_width + self.padding]


class RNNCell(BaseLayer):
    def __init__(self, parameter, mixed_precision=False):
        super().__init__()
        self.differentiable = True
        self.parameter["weight_x"] = parameter["weight_x"]
        self.parameter["weight_h"] = parameter["weight_h"]
        self.parameter["bias"] = parameter["bias"]
        self.dx = None
        self.dwx = None
        self.dwh = None
        self.db = None
        self.cache = None
        self.mixed_precision = mixed_precision

    def __repr__(self):
        return "RNNCell"

    def __call__(self, *arg):
        result = self._forward(*arg)
        return result

    def _forward(self, x, h_t_prev):
        # Mixed Precision Training
        if self.mixed_precision:
            # (batch_size, hidden_size) x (hidden_size, hidden_size) + (batch_size, input_dim) x (input_dim, hidden_size)
            # => (batch_size, hidden_size)
            temp_t = np.matmul(h_t_prev, self.parameter["weight_h"].astype(np.float16)) + np.matmul(x, self.parameter[
                "weight_x"].astype(np.float16)) + self.parameter["bias"].astype(np.float16)
            result_t = np.tanh(temp_t)
            # self.cache에 현재 timestep에서의 input, 이전 timestep에서의 hidden state, 현재 timestep에서의 output 저장
            self.cache = x, h_t_prev, result_t

        else:
            # (batch_size, hidden_size) x (hidden_size, hidden_size) + (batch_size, input_dim) x (input_dim, hidden_size)
            # => (batch_size, hidden_size)
            temp_t = np.matmul(h_t_prev, self.parameter["weight_h"]) + np.matmul(x, self.parameter["weight_x"]) + \
                     self.parameter["bias"]
            result_t = np.tanh(temp_t)
            # self.cache에 현재 timestep에서의 input, 이전 timestep에서의 hidden state, 현재 timestep에서의 output 저장
            self.cache = x, h_t_prev, result_t

        # (batch_size, hidden_size)
        return result_t

    def _backward(self, input):
        # Mixed Precision Training
        if self.mixed_precision:
            # self.cache에 저장된 현재 timestep에서의 input, 이전 timestep에서의 hidden state, 현재 timestep에서의 output 불러오기
            x, h_t_prev, result_t = self.cache
            # dtanh = 1 - tanh(x)^2
            dtanh = input * (1 - result_t ** 2)
            self.db = np.sum(dtanh, axis=0)
            self.dwh = np.matmul(h_t_prev.T, dtanh)
            self.dwx = np.matmul(x.T, dtanh)
            h_result = np.matmul(dtanh, self.parameter["weight_h"].T.astype(np.float16))
            x_result = np.matmul(dtanh, self.parameter["weight_x"].T.astype(np.float16))
            self.dx = x_result

        else:
            # self.cache에 저장된 현재 timestep에서의 input, 이전 timestep에서의 hidden state, 현재 timestep에서의 output 불러오기
            x, h_t_prev, result_t = self.cache
            # dtanh = 1 - tanh(x)^2
            dtanh = input * (1 - result_t ** 2)
            self.db = np.sum(dtanh, axis=0)
            self.dwh = np.matmul(h_t_prev.T, dtanh)
            self.dwx = np.matmul(x.T, dtanh)
            h_result = np.matmul(dtanh, self.parameter["weight_h"].T)
            x_result = np.matmul(dtanh, self.parameter["weight_x"].T)
            self.dx = x_result

        return x_result, h_result

    def _get_gradient(self):
        dx = self.dx
        dwx = self.dwx
        dwh = self.dwh
        db = self.db

        return dx, dwx, dwh, db


class RNNLayer(BaseLayer):
    def __init__(self, input_size, hidden_size, n_layers=1, bidirectional=True, stateful=True, initialize="He"):
        super().__init__()
        self.differentiable = True
        self.input_size = input_size
        self.hidden_size = hidden_size

        if initialize == "He":
            self.parameter["weight_x"] = np.random.randn(input_size, hidden_size).astype(np.float32) * (
                np.sqrt(2 / input_size))
            self.parameter["weight_h"] = np.random.randn(hidden_size, hidden_size).astype(np.float32) * (
                np.sqrt(2 / hidden_size))
            self.parameter["bias"] = np.zeros(hidden_size).astype(np.float32)

        elif initialize == "Xavier":
            self.parameter["weight_x"] = np.random.randn(input_size, hidden_size).astype(np.float32) * np.sqrt(
                1 / input_size)
            self.parameter["weight_h"] = np.random.randn(hidden_size, hidden_size).astype(np.float32) * np.sqrt(
                1 / hidden_size)
            self.parameter["bias"] = np.zeros(hidden_size).astype(np.float32)

        elif initialize == "None":
            self.parameter["weight_x"] = 0.01 * np.random.randn(input_size, hidden_size).astype(np.float32)
            self.parameter["weight_h"] = 0.01 * np.random.randn(hidden_size, hidden_size).astype(np.float32)
            self.parameter["bias"] = np.zeros(hidden_size).astype(np.float32)

        self.h = None
        self.dh = None
        self.layer = None
        self.stateful = stateful
        self.dx = None
        self.dwx = None
        self.dwh = None
        self.db = None
        self.temp = None

    def __repr__(self):
        return "RNNLayer"

    def __call__(self, arg):
        result = self._forward(arg)
        return result

    def _forward(self, x):
        # Mixed Precision Training
        if self.mixed_precision:
            wx, wh, b = self.parameter["weight_x"].astype(np.float16), self.parameter["weight_h"].astype(np.float16), \
                        self.parameter["bias"].astype(np.float16)
            batch_size, n_timestep, input_dim = x.shape
            input_dim, hidden_size = wx.shape

            hidden_state = np.empty((batch_size, n_timestep, hidden_size), dtype="f")
            self.layer = []

            if not self.stateful or self.h is None:
                self.h = np.zeros((batch_size, hidden_size)).astype(np.float32)

            for timestep in range(n_timestep):
                rnn_cell = RNNCell(self.parameter, mixed_precision=True)
                # self.h : (batch_size, 1, hidden_size)
                self.h = rnn_cell(x[:, timestep, :], self.h)
                hidden_state[:, timestep, :] = self.h
                self.layer.append(rnn_cell)

        else:
            wx, wh, b = self.parameter["weight_x"], self.parameter["weight_h"], self.parameter["bias"]
            batch_size, n_timestep, input_dim = x.shape
            input_dim, hidden_size = wx.shape

            hidden_state = np.empty((batch_size, n_timestep, hidden_size), dtype="f")
            self.layer = []

            if not self.stateful or self.h is None:
                self.h = np.zeros((batch_size, hidden_size)).astype(np.float32)

            for timestep in range(n_timestep):
                rnn_cell = RNNCell(self.parameter)
                # self.h : (batch_size, 1, hidden_size)
                self.h = rnn_cell(x[:, timestep, :], self.h)
                hidden_state[:, timestep, :] = self.h
                self.layer.append(rnn_cell)

        return hidden_state

    def _backward(self, dh):
        # Mixed Precision Training
        if self.mixed_precision:
            wx, wh, b = self.parameter["weight_x"].astype(np.float16), self.parameter["weight_h"].astype(np.float16), \
                        self.parameter["bias"].astype(np.float16)
            batch_size, n_timestep, hidden_size = dh.shape
            input_dim, hidden_size = wx.shape

            dx = np.empty((batch_size, n_timestep, input_dim), dtype="f")

            dwx = np.zeros_like(wx).astype(np.float32)
            dwh = np.zeros_like(wh).astype(np.float32)
            db = np.zeros_like(b).astype(np.float32)

            dh_t = 0

            for timestep in reversed(range(n_timestep)):
                rnn_cell = self.layer[timestep]
                dx_t, dh_t = rnn_cell._backward(dh[:, timestep, :] + dh_t)
                dx[:, timestep, :] = dx_t

                _, temp_dwx, temp_dwh, temp_db = rnn_cell._get_gradient()
                dwx += temp_dwx
                dwh += temp_dwh
                db += temp_db

            self.dwx = dwx
            self.dwh = dwh
            self.db = db
            self.dh = dh_t

        else:
            wx, wh, b = self.parameter["weight_x"], self.parameter["weight_h"], self.parameter["bias"]
            batch_size, n_timestep, hidden_size = dh.shape
            input_dim, hidden_size = wx.shape

            dx = np.empty((batch_size, n_timestep, input_dim), dtype="f")

            dwx = np.zeros_like(wx).astype(np.float32)
            dwh = np.zeros_like(wh).astype(np.float32)
            db = np.zeros_like(b).astype(np.float32)

            dh_t = 0

            for timestep in reversed(range(n_timestep)):
                rnn_cell = self.layer[timestep]
                dx_t, dh_t = rnn_cell._backward(dh[:, timestep, :] + dh_t)
                dx[:, timestep, :] = dx_t

                _, temp_dwx, temp_dwh, temp_db = rnn_cell._get_gradient()
                dwx += temp_dwx
                dwh += temp_dwh
                db += temp_db

            self.dwx = dwx
            self.dwh = dwh
            self.db = db
            self.dh = dh_t

        # input으로 backpropagation result 전달
        return dx

    def get_gradient(self) -> OrderedDict:
        """Layer의 gradient를 return

        Returns:
            OrderedDict: Layer의 gradient
        """
        grad = OrderedDict()
        grad["dwx"] = self.dwx.astype(np.float32)
        grad["dwh"] = self.dwh.astype(np.float32)
        grad["db"] = self.db.astype(np.float32)

        return grad

    def load_state(self, h):
        self.h = h

    def reset_state(self):
        self.h = None

    def _fp16_grad(self):
        pass


class LSTMCell(BaseLayer):
    def __init__(self, parameter, mixed_precision=False):
        super().__init__()
        self.differentiable = True
        self.parameter["weight_x"] = parameter["weight_x"]
        self.parameter["weight_h"] = parameter["weight_h"]
        self.parameter["bias"] = parameter["bias"]
        self.cache = None
        self.dx = None
        self.dwx = None
        self.dwh = None
        self.db = None
        self.mixed_precision = mixed_precision

    def __call__(self, *args):
        result = self._forward(*args)
        return result

    def _forward(self, x, h_t_prev, c_t_prev):
        # Mixed Precision Training
        if self.mixed_precision:
            wx, wh, b = self.parameter["weight_x"].astype(np.float16), self.parameter["weight_h"].astype(np.float16), \
                        self.parameter["bias"].astype(np.float16)
            batch_size, hidden_size = h_t_prev.shape

            a = np.matmul(x, wx) + np.matmul(h_t_prev, wh) + b
            f_temp = a[:, :hidden_size]
            g_temp = a[:, hidden_size:hidden_size * 2]
            i_temp = a[:, hidden_size * 2:hidden_size * 3]
            o_temp = a[:, hidden_size * 3:]

            f_result = sigmoid(f_temp).astype(np.float32)
            g_result = np.tanh(g_temp).astype(np.float32)
            i_result = sigmoid(i_temp).astype(np.float32)
            o_result = sigmoid(o_temp).astype(np.float32)

            c_t = f_result * c_t_prev + g_result * i_result
            h_t = o_result * np.tanh(c_t)

            self.cache = (x, h_t_prev, c_t_prev, f_result, g_result, i_result, o_result, c_t)
        else:
            wx, wh, b = self.parameter["weight_x"], self.parameter["weight_h"], self.parameter["bias"]
            batch_size, hidden_size = h_t_prev.shape

            a = np.matmul(x, wx) + np.matmul(h_t_prev, wh) + b
            f_temp = a[:, :hidden_size]
            g_temp = a[:, hidden_size:hidden_size * 2]
            i_temp = a[:, hidden_size * 2:hidden_size * 3]
            o_temp = a[:, hidden_size * 3:]

            f_result = sigmoid(f_temp)
            g_result = np.tanh(g_temp)
            i_result = sigmoid(i_temp)
            o_result = sigmoid(o_temp)

            c_t = f_result * c_t_prev + g_result * i_result
            h_t = o_result * np.tanh(c_t)

            self.cache = (x, h_t_prev, c_t_prev, f_result, g_result, i_result, o_result, c_t)

        return h_t, c_t

    def _backward(self, dh_t_next, dc_t_next):
        # Mixed Precision Training
        if self.mixed_precision:

            wx, wh, b = self.parameter["weight_x"].astype(np.float16), self.parameter["weight_h"].astype(np.float16), \
                        self.parameter["bias"].astype(np.float16)
            x, h_t_prev, c_t_prev, f_result, g_result, i_result, o_result, c_t = self.cache

            tanh_c_t = np.tanh(c_t)
            ds = dc_t_next + (dh_t_next * o_result) * (1 - tanh_c_t ** 2)
            do = dh_t_next * tanh_c_t
            di = ds * g_result
            dg = ds * i_result
            df = ds * c_t_prev

            ddo = do * o_result * (1 - o_result)
            ddi = di * i_result * (1 - i_result)
            ddg = dg * g_result * (1 - g_result ** 2)
            ddf = df * f_result * (1 - f_result)

            da = np.hstack((ddf, ddg, ddi, ddo))

            dc_t = ds * f_result

            self.dwx = np.matmul(x.T, da)
            self.dwh = np.matmul(h_t_prev.T, da)
            self.db = da.sum(axis=0)

            dx = np.matmul(da, wx.T)
            self.dx = dx

            dh_t = np.matmul(da, wh.T)

        else:
            wx, wh, b = self.parameter["weight_x"], self.parameter["weight_h"], self.parameter["bias"]
            x, h_t_prev, c_t_prev, f_result, g_result, i_result, o_result, c_t = self.cache

            tanh_c_t = np.tanh(c_t)
            ds = dc_t_next + (dh_t_next * o_result) * (1 - tanh_c_t ** 2)
            do = dh_t_next * tanh_c_t
            di = ds * g_result
            dg = ds * i_result
            df = ds * c_t_prev

            ddo = do * o_result * (1 - o_result)
            ddi = di * i_result * (1 - i_result)
            ddg = dg * g_result * (1 - g_result ** 2)
            ddf = df * f_result * (1 - f_result)

            da = np.hstack((ddf, ddg, ddi, ddo))

            dc_t = ds * f_result

            self.dwx = np.matmul(x.T, da)
            self.dwh = np.matmul(h_t_prev.T, da)
            self.db = da.sum(axis=0)

            dx = np.matmul(da, wx.T)
            self.dx = dx

            dh_t = np.matmul(da, wh.T)

        return dx, dh_t, dc_t

    def _get_gradient(self):
        dx = self.dx
        dwx = self.dwx
        dwh = self.dwh
        db = self.db

        return dx, dwx, dwh, db


class LSTMLayer(BaseLayer):
    def __init__(self, input_size, hidden_size, n_layers=1, bidirectional=True, stateful=True, initialize="He"):
        super().__init__()
        self.differentiable = True
        self.input_size = input_size
        self.hidden_size = hidden_size

        if initialize == "He":
            self.parameter["weight_x"] = np.random.randn(input_size, 4 * hidden_size).astype(np.float32) * (
                np.sqrt(2 / input_size))
            self.parameter["weight_h"] = np.random.randn(hidden_size, 4 * hidden_size).astype(np.float32) * (
                np.sqrt(2 / hidden_size))
            self.parameter["bias"] = np.zeros(4 * hidden_size).astype(np.float32)

        elif initialize == "Xavier":
            self.parameter["weight_x"] = np.random.randn(input_size, 4 * hidden_size).astype(np.float32) * np.sqrt(
                1 / input_size)
            self.parameter["weight_h"] = np.random.randn(hidden_size, 4 * hidden_size).astype(np.float32) * np.sqrt(
                1 / hidden_size)
            self.parameter["bias"] = np.zeros(4 * hidden_size).astype(np.float32)

        elif initialize == "None":
            self.parameter["weight_x"] = 0.01 * np.random.randn(input_size, 4 * hidden_size).astype(np.float32)
            self.parameter["weight_h"] = 0.01 * np.random.randn(hidden_size, 4 * hidden_size).astype(np.float32)
            self.parameter["bias"] = np.zeros(4 * hidden_size).astype(np.float32)

        self.h = None
        self.c = None
        self.dh = None
        self.layer = None
        self.stateful = stateful
        self.dx = None
        self.dwx = None
        self.dwh = None
        self.db = None
        self.temp = None

    def __repr__(self):
        return "LSTMLayer"

    def __call__(self, *args):
        result = self._forward(*args)
        return result

    def _forward(self, x):
        # Mixed Precision Training
        if self.mixed_precision:
            wx, wh, b = self.parameter["weight_x"].astype(np.float16), self.parameter["weight_h"].astype(np.float16), \
                        self.parameter["bias"].astype(np.float16)
            batch_size, n_timestep, input_dim = x.shape

            hidden_size = wh.shape[0]

            self.layer = []
            hidden_state = np.empty((batch_size, n_timestep, hidden_size), dtype="f")

            if not self.stateful or self.h is None:
                self.h = np.zeros((batch_size, hidden_size)).astype(np.float32)
            if not self.stateful or self.c is None:
                self.c = np.zeros((batch_size, hidden_size)).astype(np.float32)

            for timestep in range(n_timestep):
                lstm_cell = LSTMCell(self.parameter, mixed_precision=True)
                self.h, self.c = lstm_cell(x[:, timestep, :], self.h, self.c)
                hidden_state[:, timestep, :] = self.h

                self.layer.append(lstm_cell)

        else:
            wx, wh, b = self.parameter["weight_x"], self.parameter["weight_h"], self.parameter["bias"]
            batch_size, n_timestep, input_dim = x.shape
            hidden_size = wh.shape[0]

            self.layer = []
            hidden_state = np.empty((batch_size, n_timestep, hidden_size), dtype="f")

            if not self.stateful or self.h is None:
                self.h = np.zeros((batch_size, hidden_size)).astype(np.float32)
            if not self.stateful or self.c is None:
                self.c = np.zeros((batch_size, hidden_size)).astype(np.float32)

            for timestep in range(n_timestep):
                lstm_cell = LSTMCell(self.parameter)
                self.h, self.c = lstm_cell(x[:, timestep, :], self.h, self.c)
                hidden_state[:, timestep, :] = self.h

                self.layer.append(lstm_cell)

        return hidden_state

    def _backward(self, dh):
        # Mixed Precision Training
        if self.mixed_precision:
            wx, wh, b = self.parameter["weight_x"].astype(np.float16), self.parameter["weight_h"].astype(np.float16), \
                        self.parameter["bias"].astype(np.float16)
            batch_size, n_timestep, hidden_size = dh.shape
            input_dim = wx.shape[0]

            dx = np.empty((batch_size, n_timestep, input_dim), dtype="f")
            dh_t, dc_t = 0, 0

            dwx = np.zeros_like(wx).astype(np.float32)
            dwh = np.zeros_like(wh).astype(np.float32)
            db = np.zeros_like(b).astype(np.float32)

            for timestep in reversed(range(n_timestep)):
                lstm_cell = self.layer[timestep]
                dx_t, dh_t, dc_t = lstm_cell._backward(dh[:, timestep, :] + dh_t, dc_t)
                dx[:, timestep, :] = dx_t

                _, temp_dwx, temp_dwh, temp_db = lstm_cell._get_gradient()
                dwx += temp_dwx
                dwh += temp_dwh
                db += temp_db

            self.dwx = dwx
            self.dwh = dwh
            self.db = db
            self.dh = dh_t

        else:
            wx, wh, b = self.parameter["weight_x"], self.parameter["weight_h"], self.parameter["bias"]
            batch_size, n_timestep, hidden_size = dh.shape
            input_dim = wx.shape[0]

            dx = np.empty((batch_size, n_timestep, input_dim), dtype="f")
            dh_t, dc_t = 0, 0

            dwx = np.zeros_like(wx).astype(np.float32)
            dwh = np.zeros_like(wh).astype(np.float32)
            db = np.zeros_like(b).astype(np.float32)

            for timestep in reversed(range(n_timestep)):
                lstm_cell = self.layer[timestep]
                dx_t, dh_t, dc_t = lstm_cell._backward(dh[:, timestep, :] + dh_t, dc_t)
                dx[:, timestep, :] = dx_t

                _, temp_dwx, temp_dwh, temp_db = lstm_cell._get_gradient()
                dwx += temp_dwx
                dwh += temp_dwh
                db += temp_db

            self.dwx = dwx
            self.dwh = dwh
            self.db = db
            self.dh = dh_t

        return dx

    def get_gradient(self):
        grad = OrderedDict()
        grad["dwx"] = self.dwx.astype(np.float32)
        grad["dwh"] = self.dwh.astype(np.float32)
        grad["db"] = self.db.astype(np.float32)

        return grad

    def load_state(self, h, c=None):
        self.h, self.c = h, c

    def reset_state(self):
        self.h, self.c = None, None

    def _fp16_grad(self):
        pass


class BatchNorm1D(BaseLayer):
    def __init__(self, num_features, eps=1e-05, momentum=0.9):
        super().__init__()
        self.changeability = True
        self.differentiable = True
        self.train_fig = True
        self.num_features = num_features

        self.parameter["gamma"] = np.ones(num_features).astype(np.float32)
        self.parameter["beta"] = np.zeros(num_features).astype(np.float32)

        self.eps = eps
        self.momentum = momentum

        self.input_reshaped = None
        self.input_shape = None
        self.mean = None
        self.var = None
        self.batch_size = None
        self.xc = None
        self.std = None
        self.dgamma = None
        self.dbeta = None

    def __call__(self, *args):
        result = self._forward(*args)
        return result

    def __repr__(self):
        return "BatchNormLayer"

    def _forward(self, x):
        self.input_shape = x.shape
        if self.input_reshaped == None:
            self._check_input(x)

        if x.ndim != 2:
            batch_size, n_input_channel, input_length, = x.shape
            x = x.reshape(batch_size, -1)

        out = self.__forward(x)

        return out.reshape(*self.input_shape)

    def __forward(self, x):
        # Mixed Precision Training
        if self.mixed_precision:
            if self.mean is None:
                batch_size, input_dim = x.shape
                self.mean = np.zeros(input_dim).astype(np.float32)
                self.var = np.zeros(input_dim).astype(np.float32)

            if self.train_fig:
                mu = x.mean(axis=0)
                xc = x - mu
                var = np.mean(xc ** 2, axis=0)
                std = np.sqrt(var + self.eps)
                xn = xc / std

                self.batch_size = x.shape[0]
                self.xc = xc
                self.xn = xn
                self.std = std
                self.mean = self.momentum * self.mean + (1 - self.momentum) * mu
                self.var = self.momentum * self.var + (1 - self.momentum) * var
            else:
                xc = x - self.mean
                xn = xc / ((np.sqrt(self.var + self.eps)))

            out = self.parameter["gamma"].astype(np.float16) * xn + self.parameter["beta"].astype(np.float16)

        else:
            if self.mean is None:
                batch_size, input_dim = x.shape
                self.mean = np.zeros(input_dim).astype(np.float32)
                self.var = np.zeros(input_dim).astype(np.float32)

            if self.train_fig:
                mu = x.mean(axis=0)
                xc = x - mu
                var = np.mean(xc ** 2, axis=0)
                std = np.sqrt(var + self.eps)
                xn = xc / std

                self.batch_size = x.shape[0]
                self.xc = xc
                self.xn = xn
                self.std = std
                self.mean = self.momentum * self.mean + (1 - self.momentum) * mu
                self.var = self.momentum * self.var + (1 - self.momentum) * var
            else:
                xc = x - self.mean
                xn = xc / ((np.sqrt(self.var + self.eps)))

            out = self.parameter["gamma"] * xn + self.parameter["beta"]

        return out

    def _backward(self, dout):
        if dout.ndim != 2:
            N, C, H, W = dout.shape
            dout = dout.reshape(N, -1)

        dx = self.__backward(dout)

        dx = dx.reshape(*self.input_shape)
        return dx

    def __backward(self, dout):
        # Mixed Precision Training
        if self.mixed_precision:
            dbeta = dout.sum(axis=0)
            dgamma = np.sum(self.xn * dout, axis=0)
            dxn = self.parameter["gamma"].astype(np.float16) * dout
            dxc = dxn / self.std
            dstd = -np.sum((dxn * self.xc) / (self.std * self.std), axis=0)
            dvar = 0.5 * dstd / self.std
            dxc += (2.0 / self.batch_size) * self.xc * dvar
            dmu = np.sum(dxc, axis=0)
            dx = dxc - dmu / self.batch_size

            self.dgamma = dgamma
            self.dbeta = dbeta

        else:
            dbeta = dout.sum(axis=0)
            dgamma = np.sum(self.xn * dout, axis=0)
            dxn = self.parameter["gamma"] * dout
            dxc = dxn / self.std
            dstd = -np.sum((dxn * self.xc) / (self.std * self.std), axis=0)
            dvar = 0.5 * dstd / self.std
            dxc += (2.0 / self.batch_size) * self.xc * dvar
            dmu = np.sum(dxc, axis=0)
            dx = dxc - dmu / self.batch_size

            self.dgamma = dgamma
            self.dbeta = dbeta

        return dx

    def _check_input(self, x):
        if x.ndim != 2:
            batch_size, n_input_channel, input_length = self.input_shape
            self.parameter["gamma"] = np.ones(self.num_features * input_length).astype(np.float32)
            self.parameter["beta"] = np.zeros(self.num_features * input_length).astype(np.float32)
            self.input_reshaped = True
        else:
            self.input_reshaped = False

    def get_gradient(self) -> OrderedDict:
        """Layer의 gradient를 return

        Returns:
            OrderedDict: Layer의 gradient
        """
        grad = OrderedDict()
        grad["dgamma"] = self.dgamma.astype(np.float32)
        grad["dbeta"] = self.dbeta.astype(np.float32)

        return grad

    def train_state(self):
        self.train_fig = True

    def eval_state(self):
        self.train_fig = False


class BatchNorm2D(BaseLayer):
    def __init__(self, num_features, eps=1e-05, momentum=0.9):
        super().__init__()
        self.changeability = True
        self.differentiable = True
        self.train_fig = True
        self.num_features = num_features
        self.parameter["gamma"] = None
        self.parameter["beta"] = None
        self.eps = eps
        self.momentum = momentum

        self.input_reshaped = None
        self.input_shape = None
        self.mean = None
        self.var = None
        self.batch_size = None
        self.xc = None
        self.std = None
        self.dgamma = None
        self.dbeta = None

    def __call__(self, args):
        result = self._forward(args)
        return result

    def __repr__(self):
        return "BatchNormLayer"

    def _forward(self, x):
        self.input_shape = x.shape
        # print(x.shape)
        if self.input_reshaped == None:
            self._initialize_param()

        batch_size, n_input_channel, input_height, input_width = x.shape
        x = x.reshape(batch_size, -1)

        out = self.__forward(x)

        return out.reshape(*self.input_shape)

    def __forward(self, x):
        # Mixed Precision Training
        if self.mixed_precision:
            if self.mean is None:
                batch_size, input_dim = x.shape
                self.mean = np.zeros(input_dim).astype(np.float32)
                self.var = np.zeros(input_dim).astype(np.float32)

            if self.train_fig:
                mu = x.mean(axis=0)
                xc = x - mu
                var = np.mean(xc ** 2, axis=0)
                std = np.sqrt(var + self.eps)
                xn = xc / std

                self.batch_size = x.shape[0]
                self.xc = xc
                self.xn = xn
                self.std = std
                self.mean = self.momentum * self.mean + (1 - self.momentum) * mu
                self.var = self.momentum * self.var + (1 - self.momentum) * var
            else:
                xc = x - self.mean
                xn = xc / ((np.sqrt(self.var + self.eps)))

            out = self.parameter["gamma"].astype(np.float16) * xn + self.parameter["beta"].astype(np.float16)

        else:
            if self.mean is None:
                batch_size, input_dim = x.shape
                self.mean = np.zeros(input_dim).astype(np.float32)
                self.var = np.zeros(input_dim).astype(np.float32)

            if self.train_fig:
                mu = x.mean(axis=0)
                xc = x - mu
                var = np.mean(xc ** 2, axis=0)
                std = np.sqrt(var + self.eps)
                xn = xc / std

                self.batch_size = x.shape[0]
                self.xc = xc
                self.xn = xn
                self.std = std
                self.mean = self.momentum * self.mean + (1 - self.momentum) * mu
                self.var = self.momentum * self.var + (1 - self.momentum) * var
            else:
                xc = x - self.mean
                xn = xc / ((np.sqrt(self.var + self.eps)))

            out = self.parameter["gamma"] * xn + self.parameter["beta"]

        return out

    def _backward(self, dout):
        N, C, H, W = dout.shape
        dout = dout.reshape(N, -1)

        dx = self.__backward(dout)

        dx = dx.reshape(*self.input_shape)
        return dx

    def __backward(self, dout):
        # Mixed Precision Training
        if self.mixed_precision:
            dbeta = dout.sum(axis=0)
            dgamma = np.sum(self.xn * dout, axis=0)
            dxn = self.parameter["gamma"].astype(np.float16) * dout
            dxc = dxn / self.std
            dstd = -np.sum((dxn * self.xc) / (self.std * self.std), axis=0)
            dvar = 0.5 * dstd / self.std
            dxc += (2.0 / self.batch_size) * self.xc * dvar
            dmu = np.sum(dxc, axis=0)
            dx = dxc - dmu / self.batch_size

            self.dgamma = dgamma
            self.dbeta = dbeta

        else:
            dbeta = dout.sum(axis=0)
            dgamma = np.sum(self.xn * dout, axis=0)
            dxn = self.parameter["gamma"] * dout
            dxc = dxn / self.std
            dstd = -np.sum((dxn * self.xc) / (self.std * self.std), axis=0)
            dvar = 0.5 * dstd / self.std
            dxc += (2.0 / self.batch_size) * self.xc * dvar
            dmu = np.sum(dxc, axis=0)
            dx = dxc - dmu / self.batch_size

            self.dgamma = dgamma
            self.dbeta = dbeta

        return dx

    def _initialize_param(self):
        batch_size, n_input_channel, input_height, input_width = self.input_shape
        self.parameter["gamma"] = np.ones(self.num_features * input_height * input_width).astype(np.float32)
        self.parameter["beta"] = np.zeros(self.num_features * input_height * input_width).astype(np.float32)
        self.input_reshaped = True

    def get_gradient(self) -> OrderedDict:
        """Layer의 gradient를 return

        Returns:
            OrderedDict: Layer의 gradient
        """
        grad = OrderedDict()
        grad["dgamma"] = self.dgamma.astype(np.float32)
        grad["dbeta"] = self.dbeta.astype(np.float32)

        return grad

    def train_state(self):
        self.train_fig = True

    def eval_state(self):
        self.train_fig = False

    def _fp16_grad(self):
        pass


class LayerNorm(BaseLayer):
    def __init__(self, input_shape, eps=1e-05, momentum=0.9):
        super().__init__()
        self.changeability = False
        self.differentiable = True
        if isinstance(input_shape, int):
            input_shape = (input_shape,)
        self.input_shape = tuple(input_shape)

        self.parameter["gamma"] = np.ones(self.input_shape).astype(np.float32)
        self.parameter["beta"] = np.zeros(self.input_shape).astype(np.float32)

        self.eps = eps
        self.xc = None
        self.std = None
        self.dgamma = None
        self.dbeta = None

    def __call__(self, *args):
        result = self._forward(*args)
        return result

    def __repr__(self):
        return "LayerNormLayer"

    def _forward(self, x):
        # Mixed Precision Training
        if self.mixed_precision:
            dims = tuple([-(i + 1) for i in range(len(x.shape[1:]))])
            # mu.shape : (batch_size, 1, 1)
            mu = x.mean(axis=dims, keepdims=True)
            xc = x - mu
            var = np.mean(xc ** 2, axis=dims, keepdims=True)
            std = np.sqrt(var + self.eps)
            xn = xc / std

            self.xc = xc
            self.xn = xn
            self.std = std

            out = self.parameter["gamma"].astype(np.float16) * xn + self.parameter["beta"].astype(np.float16)

        else:
            dims = tuple([-(i + 1) for i in range(len(x.shape[1:]))])
            # mu.shape : (batch_size, 1, 1)
            mu = x.mean(axis=dims, keepdims=True)
            xc = x - mu
            var = np.mean(xc ** 2, axis=dims, keepdims=True)
            std = np.sqrt(var + self.eps)
            xn = xc / std

            self.xc = xc
            self.xn = xn
            self.std = std

            out = self.parameter["gamma"] * xn + self.parameter["beta"]

        return out

    def _backward(self, dout):
        # Mixed Precision Training
        if self.mixed_precision:
            dbeta = dout.sum(axis=0)
            dbeta = dbeta.sum(axis=0)
            dgamma = np.sum(self.xn * dout, axis=0)
            dgamma = dgamma.sum(axis=0)
            dxn = self.parameter["gamma"].astype(np.float16) * dout
            dxc = dxn / self.std
            dstd = -np.sum((dxn * self.xc) / (self.std * self.std), axis=0)
            dvar = 0.5 * dstd / self.std
            dxc += (2.0 / self.xc[0].size) * self.xc * dvar

            dmu = np.sum(dxc, axis=0)
            dx = dxc - dmu / self.xc[0].size

            self.dgamma = dgamma
            self.dbeta = dbeta

        else:
            dbeta = dout.sum(axis=0)
            dgamma = np.sum(self.xn * dout, axis=0)
            dxn = self.parameter["gamma"] * dout
            dxc = dxn / self.std
            dstd = -np.sum((dxn * self.xc) / (self.std * self.std), axis=0)
            dvar = 0.5 * dstd / self.std
            dxc += (2.0 / self.xc[0].size) * self.xc * dvar

            dmu = np.sum(dxc, axis=0)
            dx = dxc - dmu / self.xc[0].size

            self.dgamma = dgamma
            self.dbeta = dbeta

        return dx

    def get_gradient(self) -> OrderedDict:
        """Layer의 gradient를 return

        Returns:
            OrderedDict: Layer의 gradient
        """
        grad = OrderedDict()
        grad["dgamma"] = self.dgamma.astype(np.float32)
        grad["dbeta"] = self.dbeta.astype(np.float32)

        return grad


class Dropout(BaseLayer):
    def __init__(self, dropout_ratio=0.5):
        super().__init__()
        self.changeability = True
        self.dropout_ratio = dropout_ratio
        self.train_fig = True
        self.mask = None

    def __repr__(self):
        return "Dropout"

    def __call__(self, args):
        result = self._forward(args)
        return result

    def _forward(self, x):
        # Mixed Precision Training
        if self.train_fig:
            fig = np.random.rand(*x.shape) > self.dropout_ratio
            scale = 1 / (1.0 - self.dropout_ratio)
            self.mask = fig.astype(np.float32) * scale
            return x * self.mask
        else:
            return x

            # if self.train_fig:
            #     self.mask = np.random.rand(*x.shape) > self.dropout_ratio
            #     return x * self.mask
            # else:
            #     return x * (1.0 - self.dropout_ratio)

    def _backward(self, dout):
        return dout * self.mask

    def train_state(self):
        self.train_fig = True

    def eval_state(self):
        self.train_fig = False


class Model(BaseModel):
    def __init__(self, *layers):
        """매개 변수에 layer 객체를 순차적으로 전달하면서 초기화
        """
        super().__init__()
        self.sequence = []
        self.grad = OrderedDict()

        self.tying_weight = False

        self.layers = layers
        self.count_dict = OrderedDict()
        temp_repr_list = []

        for layer in self.layers:
            temp_repr_list.append(repr(layer))
        repr_set = set(temp_repr_list)

        # initialize count_dict
        for rep in repr_set:
            self.count_dict[rep] = 1

        for layer in self.layers:
            self.network[f"{repr(layer)}{self.count_dict[repr(layer)]}"] = layer
            self.sequence.append(f"{repr(layer)}{self.count_dict[repr(layer)]}")
            self.count_dict[repr(layer)] += 1

    def __call__(self, arg):
        result = self.forward(arg)
        return result

    def __str__(self):
        structure = ""
        string_list = []
        for i, layer_name in enumerate(self.sequence):
            layer = self.network[layer_name]

            if layer.differentiable:
                if repr(layer) == "RNNLayer":
                    shape = layer.parameter["weight_x"].shape
                    string_list.append(f"{i}. {layer_name} : {layer} {shape} \n")

                elif repr(layer) == "LSTMLayer":
                    shape = layer.parameter["weight_x"].shape
                    lstm_in, lstm_out = shape
                    lstm_out /= 4
                    shape = (lstm_in, int(lstm_out))
                    string_list.append(f"{i}. {layer_name} : {layer} {shape} \n")

                elif repr(layer) == "BatchNormLayer":
                    shape = layer.num_features
                    string_list.append(f"{i}. {layer_name} : {layer} ({shape}) \n")

                elif repr(layer) == "LayerNormLayer":
                    shape = layer.input_shape
                    string_list.append(f"{i}. {layer_name} : {layer} {shape} \n")

                else:
                    shape = layer.parameter["weight"].shape
                    string_list.append(f"{i}. {layer_name} : {layer} {shape} \n")
            else:
                if repr(layer) == "Dropout":
                    shape = layer.dropout_ratio
                    string_list.append(f"{i}. {layer_name} : {layer} ({shape}) \n")

                else:
                    string_list.append(f"{i}. {layer_name} : {layer}\n")

        structure = structure.join(string_list)
        return structure

    def forward(self, x):
        input = x
        if self.tying_weight:
            self.tying_forward()
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            y = layer(input)
            input = y

        return input

    def backward(self, loss):
        result = loss._backward()
        for layer_name in reversed(self.sequence):
            layer = self.network[layer_name]
            result = layer._backward(result)

        if self.tying_weight:
            self.tying_backward()

    def get_gradient(self):
        grad = OrderedDict()
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            if layer.differentiable:
                grad[layer_name] = layer.get_gradient()

        return grad

    def add_layer(self, *layers):
        tuple_layers = layers
        for layer in tuple_layers:
            if repr(layer) not in self.count_dict.keys():
                self.count_dict[repr(layer)] = 1

            self.layers += (layer,)
            self.network[f"{repr(layer)}{self.count_dict[repr(layer)]}"] = layer
            self.sequence.append(f"{repr(layer)}{self.count_dict[repr(layer)]}")
            self.count_dict[repr(layer)] += 1

    def train_state(self):
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            if layer.changeability:
                layer.train_state()

    def eval_state(self):
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            if layer.changeability:
                layer.eval_state()

    def reset_rnn_state(self):
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            if repr(layer) == "LSTMLayer" or repr(layer) == "RNNLayer":
                layer.reset_state()

    def weight_tying(self):
        emb_layer_name = self.sequence[0]
        final_dense_name = self.sequence[-1]

        emb_layer = self.network[emb_layer_name]
        final_dense_layer = self.network[final_dense_name]

        emb_layer.tied = True
        final_dense_layer.tying = True
        self.tying_weight = True

    def tying_forward(self):
        emb_layer_name = self.sequence[0]
        final_dense_name = self.sequence[-1]

        emb_layer = self.network[emb_layer_name]
        final_dense_layer = self.network[final_dense_name]

        emb_layer.parameter["weight"] = final_dense_layer.parameter["weight"].T

    def tying_backward(self):
        emb_layer_name = self.sequence[0]
        final_dense_name = self.sequence[-1]

        emb_layer = self.network[emb_layer_name]
        final_dense_layer = self.network[final_dense_name]

        final_dense_layer.dw += emb_layer.dw.T

    def to_cpu(self):
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            if layer.differentiable:
                layer._to_cpu()

    def to_gpu(self):
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            if layer.differentiable:
                layer._to_gpu()

    def save_params(self, fn: str = None):
        """Model의 parameter .pkl 확장자로 저장

        Args:
            fn (str): parameter를 저장할 경로 + 파일 이름 + .pkl, Defaults to None.
        """
        model_param = OrderedDict()
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            if layer.differentiable:
                param = layer._save_params()
                model_param[layer_name] = param

        if fn is None:
            fn = self.__class__.__name__ + '.pkl'

        with open(fn, 'wb') as f:
            pickle.dump(model_param, f)

    def load_params(self, fn: str = None):
        """.pkl 확장자로 저장되어있던 Model의 parameter를 가져온다

        Args:
            fn (str): parameter가 저장된 경로 + 파일 이름 + .pkl, Defaults to None.
        """
        with open(fn, 'rb') as f:
            model_param = deepcopy(pickle.load(f))

        for layer_name in self.sequence:
            layer = self.network[layer_name]
            if layer.differentiable:
                layer._load_params(model_param[layer_name])

    def mixed_precision_on(self, grad_init=True):
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            layer._mixed_precision_training()

    def mixed_precision_off(self):
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            layer._mixed_precision_training_off()







from config import GPU

if GPU:
    import cupy as np
    np.cuda.set_allocator(np.cuda.MemoryPool().malloc)

    print('\033[92m' + '-' * 60 + '\033[0m')
    print(' ' * 23 + '\033[92mGPU Mode (cupy)\033[0m')
    print('\033[92m' + '-' * 60 + '\033[0m\n')
else:
    import numpy as np




from copy import deepcopy
from collections import OrderedDict
from gpu import *
from function import *
from function_class import *
from model import *


class NegativeSamplingLoss(Model):
    def __init__(self, corpus, vocab_size, hidden_size, power=0.75, sample_size=5, initialize="He"):
        self.sample_size = sample_size
        self.sampler = NegativeSampler(corpus, power, sample_size)
        self.loss_layers = tuple([BinaryCrossEntropyLoss() for _ in range(sample_size + 1)])
        self.embed_dot_layers = tuple([EmbeddingDot(vocab_size, hidden_size) for _ in range(sample_size + 1)])
        self.layers = self.loss_layers + self.embed_dot_layers + (self.sampler,)

        self.network = OrderedDict()
        self.count_dict = OrderedDict()
        self.sequence = []
        temp_repr_list = []

        for layer in self.layers:
            temp_repr_list.append(repr(layer))
        repr_set = set(temp_repr_list)

        # initialize count_dict
        for rep in repr_set:
            self.count_dict[rep] = 1

        for layer in self.layers:
            self.network[f"{repr(layer)}{self.count_dict[repr(layer)]}"] = layer
            self.sequence.append(f"{repr(layer)}{self.count_dict[repr(layer)]}")
            self.count_dict[repr(layer)] += 1

    def forward(self, h, target):
        batch_size = target.shape[0]
        # 자주 등장하는 단어 더 추출하기
        # negative_sample = (batch_size, sample_size)
        negative_sample = self.sampler.negative_sampling(target)

        score = self.embed_dot_layers[0]._forward(h, target)
        # 정답 : 1
        correct_label = np.ones(batch_size, dtype=np.int32)
        loss = self.loss_layers[0]._forward(score, correct_label)

        # 오답 : 0
        negative_label = np.zeros(batch_size, dtype=np.int32)

        # sample_size = 몇개 sampling할건지
        for i in range(self.sample_size):
            negative_target = negative_sample[:, i]
            score = self.embed_dot_layers[1 + i]._forward(h, negative_target)
            loss += self.loss_layers[1 + i]._forward(score, negative_label)

        return loss

    def backward(self, dout=1):
        dh = 0
        for loss_layer, emb_layer in zip(self.loss_layers, self.embed_dot_layers):
            dscore = loss_layer._backward(dout)
            dh += emb_layer._backward(dscore)

        return dh


class Cbow(Model):
    def __init__(self, corpus, vocab_size, hidden_size, window_size, sample_size=5, initialize="He"):
        self.network = OrderedDict()
        self.sequence = []
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.parameter = OrderedDict()
        if initialize == "He":
            self.parameter["weight"] = np.random.randn(vocab_size, hidden_size).astype(np.float32) * (
                np.sqrt(2 / vocab_size))

        elif initialize == "Xavier":
            self.parameter["weight"] = np.random.randn(vocab_size, hidden_size).astype(np.float32) * np.sqrt(
                1 / vocab_size)

        elif initialize == "None":
            self.parameter["weight"] = 0.01 * np.random.randn(vocab_size, hidden_size).astype(np.float32)

        elif isinstance(initialize, int):
            self.parameter["weight"] = (1 / initialize) * np.random.randn(vocab_size, hidden_size).astype(np.float32)

        else:
            raise ValueError("'initialize' must be 'He' or 'Xavier' or 'None' or integer")

        self.layers = ()
        self.input_layers = ()
        for i in range(window_size * 2):
            self.input_layers += (Embedding(self.parameter),)

        self.loss = NegativeSamplingLoss(corpus, vocab_size, hidden_size, sample_size=sample_size)
        self.layers += self.input_layers
        self.layers += self.loss.layers

        self.count_dict = OrderedDict()
        temp_repr_list = []

        for layer in self.layers:
            temp_repr_list.append(repr(layer))
        repr_set = set(temp_repr_list)

        # initialize count_dict
        for rep in repr_set:
            self.count_dict[rep] = 1

        for layer in self.layers:
            self.network[f"{repr(layer)}{self.count_dict[repr(layer)]}"] = layer
            self.sequence.append(f"{repr(layer)}{self.count_dict[repr(layer)]}")
            self.count_dict[repr(layer)] += 1

        self.word_vector = self.parameter["weight"]

    def forward(self, context, target):
        h = 0
        for i, layer in enumerate(self.input_layers):
            h += layer._forward(context[:, i])

        h *= 1 / len(self.input_layers)
        loss = self.loss.forward(h, target)
        return loss

    def backward(self, dout=1):
        dout = self.loss.backward(dout)
        dout *= 1 / len(self.input_layers)
        for layer in self.input_layers:
            layer._backward(dout)


class SkipGram(Model):
    def __init__(self, corpus, vocab_size, hidden_size, window_size, sample_size=5, initialize="He"):
        self.network = OrderedDict()
        self.sequence = []
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.parameter = OrderedDict()
        if initialize == "He":
            self.parameter["weight"] = np.random.randn(vocab_size, hidden_size).astype(np.float32) * (
                np.sqrt(2 / vocab_size))

        elif initialize == "Xavier":
            self.parameter["weight"] = np.random.randn(vocab_size, hidden_size).astype(np.float32) * np.sqrt(
                1 / vocab_size)

        elif initialize == "None":
            self.parameter["weight"] = 0.01 * np.random.randn(vocab_size, hidden_size).astype(np.float32)

        elif isinstance(initialize, int):
            self.parameter["weight"] = (1 / initialize) * np.random.randn(vocab_size, hidden_size).astype(np.float32)

        else:
            raise ValueError("'initialize' must be 'He' or 'Xavier' or 'None' or integer")

        self.layers = ()
        self.input_layer = Embedding(self.parameter)
        self.layers += (self.input_layer,)

        self.loss = ()
        for i in range(window_size * 2):
            self.loss += (NegativeSamplingLoss(corpus, vocab_size, hidden_size, sample_size=sample_size),)

        for loss in self.loss:
            self.layers += loss.layers

        self.count_dict = OrderedDict()
        temp_repr_list = []

        for layer in self.layers:
            temp_repr_list.append(repr(layer))
        repr_set = set(temp_repr_list)

        # initialize count_dict
        for rep in repr_set:
            self.count_dict[rep] = 1

        for layer in self.layers:
            self.network[f"{repr(layer)}{self.count_dict[repr(layer)]}"] = layer
            self.sequence.append(f"{repr(layer)}{self.count_dict[repr(layer)]}")
            self.count_dict[repr(layer)] += 1

        self.word_vector = self.parameter["weight"]

    def forward(self, context, target):
        h = self.input_layer._forward(target)

        loss = 0
        for i, layer in enumerate(self.loss):
            loss += layer.forward(h, context[:, i])
        return loss

    def backward(self, dout=1):
        dh = 0
        for i, layer in enumerate(self.loss):
            dh += layer.backward(dout)
        self.input_layer._backward(dh)

        return None


class LanguageModel(Model):
    def generate(self, start_id, skip_ids=None, sample_size=100):
        word_ids = [start_id]

        x = start_id
        while len(word_ids) < sample_size:
            x = np.array(x).reshape(1, 1)
            self.eval_state()
            score = self.forward(x)
            pred = softmax(score.flatten())

            sampled = np.random.choice(len(pred), size=1, p=pred)
            if (skip_ids is None) or (sampled not in skip_ids):
                x = sampled
                word_ids.append(int(x))

        return word_ids

    def get_rnn_state(self):
        states = OrderedDict()
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            if repr(layer) == "LSTMLayer":
                states[layer_name] = OrderedDict()
                states[layer_name]["h"] = deepcopy(layer.h)
                states[layer_name]["c"] = deepcopy(layer.c)

            elif repr(layer) == "RNNLayer":
                states[layer_name] = OrderedDict()
                states[layer_name]["h"] = deepcopy(layer.h)

        return states

    def set_rnn_state(self, states):
        for layer_name in self.sequence:
            layer = self.network[layer_name]
            if repr(layer) == "LSTMLayer":
                layer.h = deepcopy(states[layer_name]["h"])
                layer.c = deepcopy(states[layer_name]["c"])

            elif repr(layer) == "RNNLayer":
                layer.h = deepcopy(states[layer_name]["h"])


class Encoder(Model):
    def __init__(self, vocab_size=1000, word_vec_size=128, hidden_size=128, n_layers=1, dropout=None, *layers):
        super().__init__(*layers)
        if len(layers) == 0:
            self.add_layer(EmbeddingLayer(vocab_size=vocab_size, hidden_size=word_vec_size))
            for i in range(n_layers):
                self.add_layer(LSTMLayer(word_vec_size, hidden_size, stateful=False))
                if dropout != None:
                    self.add_layer(Dropout(dropout_ratio=dropout))

        self.hidden_state = None

    def _forward(self, x):
        hidden_state = self.forward(x)
        self.hidden_state = hidden_state

        return hidden_state[:, -1, :]

    def _backward(self, dh_t):
        dhidden_state = np.zeros_like(self.hidden_state)
        dhidden_state[:, -1, :] = dh_t

        last_layer_name = self.sequence[-1]
        last_layer = self.network[last_layer_name]
        result = last_layer._backward(dhidden_state)

        for layer_name in reversed(self.sequence[:-1]):
            layer = self.network[layer_name]
            result = layer._backward(result)

        if self.tying_weight:
            self.tying_backward()


class Decoder(Model):
    def __init__(self, vocab_size=1000, word_vec_size=128, hidden_size=128, n_layers=1, peeky=False, dropout=None,
                 *layers):
        super().__init__(*layers)
        if len(layers) == 0:
            if peeky:
                self.add_layer(EmbeddingLayer(vocab_size=vocab_size, hidden_size=word_vec_size, initialize=100))
                for i in range(n_layers):
                    self.add_layer(LSTMLayer(word_vec_size + hidden_size, hidden_size))
                    if dropout != None:
                        self.add_layer(Dropout(dropout_ratio=dropout))

                self.add_layer(DenseLayer(hidden_size + hidden_size, vocab_size))

            else:
                self.add_layer(EmbeddingLayer(vocab_size=vocab_size, hidden_size=word_vec_size, initialize=100))
                for i in range(n_layers):
                    self.add_layer(LSTMLayer(word_vec_size, hidden_size))
                    if dropout != None:
                        self.add_layer(Dropout(dropout_ratio=dropout))

                self.add_layer(DenseLayer(hidden_size, vocab_size))

        self.hidden_size = hidden_size
        self.peeky = peeky

    def _forward(self, x, h):
        if self.tying_weight:
            self.tying_forward()

        if self.peeky:
            batch_size, n_timestep = x.shape

            layer = self.network["LSTMLayer1"]
            layer.load_state(h)

            first_emb_name = self.sequence[0]
            first_emb = self.network[first_emb_name]
            result = first_emb(x)

            hidden_state = np.repeat(h, n_timestep, axis=0).reshape(batch_size, n_timestep, self.hidden_size)

            for layer_name in self.sequence[1:-1]:
                layer = self.network[layer_name]
                if repr(layer) == "LSTMLayer" or repr(layer) == "RNNLayer":
                    result = np.concatenate((hidden_state, result), axis=2)
                result = layer(result)

            result = np.concatenate((hidden_state, result), axis=2)
            final_dense_name = self.sequence[-1]
            final_dense = self.network[final_dense_name]
            result = final_dense(result)

        else:
            layer = self.network["LSTMLayer1"]
            layer.load_state(h)

            input = x

            for layer_name in self.sequence:
                layer = self.network[layer_name]
                y = layer(input)
                input = y

            result = input

        return result

    def _backward(self, loss):
        if self.peeky:
            final_dense_name = self.sequence[-1]
            final_dense = self.network[final_dense_name]
            dout = final_dense._backward(loss)
            dout, dhs0 = dout[:, :, self.hidden_size:], dout[:, :, :self.hidden_size]

            for layer_name in reversed(self.sequence[1:-1]):
                layer = self.network[layer_name]
                dout = layer._backward(dout)
                if repr(layer) == "LSTMLayer" or repr(layer) == "RNNLayer":
                    dembed, dhs1 = dout[:, :, self.hidden_size:], dout[:, :, :self.hidden_size]

            first_emb_name = self.sequence[0]
            first_emb = self.network[first_emb_name]
            first_emb._backward(dembed)
            dhs = dhs0 + dhs1

            first_lstm_dh = self.network["LSTMLayer1"].dh
            dh = first_lstm_dh + np.sum(dhs, axis=1)

        else:
            last_layer_name = self.sequence[-1]
            last_layer = self.network[last_layer_name]
            result = last_layer._backward(loss)

            for layer_name in reversed(self.sequence[:-1]):
                layer = self.network[layer_name]
                result = layer._backward(result)

            dh = 0
            dh = self.network["LSTMLayer1"].dh

        if self.tying_weight:
            self.tying_backward()

        return dh

    def generate(self, h, start_id, sample_size):
        sampled_result = []
        sample_id = start_id
        layer = self.network["LSTMLayer1"]
        layer.load_state(h)

        if self.peeky:
            peeky_h = h.reshape(1, 1, self.hidden_size)
            for i in range(sample_size):
                x = np.array(sample_id).reshape((1, 1))
                result = x

                first_emb_name = self.sequence[0]
                first_emb = self.network[first_emb_name]
                result = first_emb(x)

                for layer_name in self.sequence[1:-1]:
                    layer = self.network[layer_name]
                    if repr(layer) == "LSTMLayer" or repr(layer) == "RNNLayer":
                        result = np.concatenate((peeky_h, result), axis=2)
                    result = layer(result)

                result = np.concatenate((peeky_h, result), axis=2)
                final_dense_name = self.sequence[-1]
                final_dense = self.network[final_dense_name]
                result = final_dense(result)

                sample_id = np.argmax(result.flatten())
                sampled_result.append(int(sample_id))

        else:
            for i in range(sample_size):
                x = np.array(sample_id).reshape((1, 1))
                result = x
                for layer_name in self.sequence:
                    layer = self.network[layer_name]
                    result = layer._forward(result)

                sample_id = np.argmax(result.flatten())
                sampled_result.append(int(sample_id))

        return sampled_result


class Seq2Seq(Model):
    def __init__(self, vocab_size=1000, word_vec_size=128, hidden_size=128, n_layers=1, peeky=False, dropout=None):
        super().__init__(())
        self.encoder = Encoder(vocab_size=vocab_size,
                               word_vec_size=word_vec_size,
                               hidden_size=hidden_size,
                               n_layers=n_layers,
                               dropout=dropout)

        if peeky:
            self.decoder = Decoder(vocab_size=vocab_size,
                                   word_vec_size=word_vec_size,
                                   hidden_size=hidden_size,
                                   n_layers=n_layers,
                                   peeky=True,
                                   dropout=dropout)

        else:
            self.decoder = Decoder(vocab_size=vocab_size,
                                   word_vec_size=word_vec_size,
                                   hidden_size=hidden_size,
                                   n_layers=n_layers,
                                   dropout=dropout)

        self.network = OrderedDict()
        self.count_dict = OrderedDict()
        self.layers = self.encoder.layers + self.decoder.layers
        self.sequence = []
        temp_repr_list = []

        for layer in self.layers:
            temp_repr_list.append(repr(layer))
        repr_set = set(temp_repr_list)

        # initialize count_dict
        for rep in repr_set:
            self.count_dict[rep] = 1

        for layer in self.layers:
            self.network[f"{repr(layer)}{self.count_dict[repr(layer)]}"] = layer
            self.sequence.append(f"{repr(layer)}{self.count_dict[repr(layer)]}")
            self.count_dict[repr(layer)] += 1

    def __repr__(self) -> str:
        return "Seq2Seq"

    def __call__(self, *args):
        result = self.forward(*args)
        return result

    def forward(self, x, decoder_x):
        h = self.encoder._forward(x)
        result = self.decoder._forward(decoder_x, h)

        return result

    def backward(self, loss):
        dout = loss._backward()
        dh = self.decoder._backward(dout)
        self.encoder._backward(dh)

    def generate(self, x, start_id, sample_size):
        h = self.encoder._forward(x)
        sampled = self.decoder.generate(h,
                                        start_id=start_id,
                                        sample_size=sample_size)

        return sampled

    def weight_tying(self):
        self.decoder.weight_tying()


class WeightSum(BaseFunction):
    def __init__(self):
        super().__init__()
        self.differentiable = False
        self.changeability = False
        self.mixed_precision = False

        self.cache = None

    def _forward(self, hidden_state, weigth):
        batch_size, n_timestep, hidden_size = hidden_state.shape

        # with broadcast
        # repeated_attention : (batch_size, n_timestep, hidden_state)
        repeated_attention = weigth.reshape(batch_size, n_timestep, 1)

        # hidden_state : (batch_size, n_timestep, hidden_state)
        t = hidden_state * repeated_attention

        # hidden_size의 요소들 하나로 합쳐서 context생성
        # context : (batch_size, hidden_state)
        context = np.sum(t, axis=1)

        self.cache = (hidden_state, repeated_attention)
        return context

    def _backward(self, dcontext):
        hidden_state, repeated_attention = self.cache
        batch_size, n_timestep, hidden_size = hidden_state.shape

        dt = dcontext.reshape(batch_size, 1, hidden_size).repeat(n_timestep, axis=1)

        drepeated_attention = dt * hidden_state
        dhidden_state = dt * repeated_attention
        # repeat된거 합하기
        dweight = np.sum(drepeated_attention, axis=2)

        return dhidden_state, dweight


class AttentionWeight(BaseFunction):
    def __init__(self):
        super().__init__()
        self.softmax = Softmax()
        self.cache = None
        self.differentiable = False
        self.changeability = False
        self.mixed_precision = False

    def _forward(self, hidden_state, h_t):
        batch_size, n_timestep, hidden_size = hidden_state.shape

        # h_t : 하나의 timestep의 hidden_state
        # (batch_size, hidden_size) -> (batch_size, 1, hidden_size)
        repeated_h = h_t.reshape(batch_size, 1, hidden_size)
        # t : (batch_size, n_timestep ,hidden_size)
        t = hidden_state * repeated_h

        # score : (batch_size, n_timestep)
        score = np.sum(t, axis=2)
        # attention : (batch_size, n_timestep)
        attention = self.softmax._forward(score)

        self.cache = (hidden_state, repeated_h)
        return attention

    def _backward(self, dattention):
        hidden_state, repeated_h = self.cache
        batch_size, n_timestep, hidden_size = hidden_state.shape

        # score : (batch_size, n_timestep)
        dscore = self.softmax._backward(dattention)

        # t : (batch_size, n_timestep ,hidden_size)
        dt = dscore.reshape(batch_size, n_timestep, 1).repeat(hidden_size, axis=2)
        dhidden_state = dt * repeated_h
        drepeated_h = dt * hidden_state
        # repeat된거 합하기
        dh_t = np.sum(drepeated_h, axis=1)

        return dhidden_state, dh_t


class Attention(BaseFunction):
    def __init__(self):
        super().__init__()
        self.differentiable = False
        self.changeability = False
        self.mixed_precision = False

        self.weight_sum = WeightSum()
        self.attention_weight = AttentionWeight()
        self.attention = None

    def _forward(self, hidden_state, h_t):
        attention = self.attention_weight._forward(hidden_state, h_t)
        result = self.weight_sum._forward(hidden_state, attention)
        self.attention = attention
        return result

    def _backward(self, dout):
        dhidden_state_0, dattention = self.weight_sum._backward(dout)
        dhidden_state_1, dh_t = self.attention_weight._backward(dattention)
        dhidden_state = dhidden_state_0 + dhidden_state_1
        return dhidden_state, dh_t


class AttentionLayer(BaseLayer):
    def __init__(self):
        super().__init__()
        self.differentiable = False
        self.changeability = False
        self.mixed_precision = False
        self.attention_cells = None
        self.attention_weights = None

    def __call__(self, *args):
        result = self._forward(*args)
        return result

    def __repr__(self) -> str:
        return "AttentionLayer"

    def _forward(self, encoder_hidden_state, decoder_hidden_state):
        batch_size, n_timestep, hidden_size = decoder_hidden_state.shape
        result = np.empty_like(decoder_hidden_state).astype(np.float32)
        self.attention_cells = []
        self.attention_weights = []

        for timestep in range(n_timestep):
            cell = Attention()
            result[:, timestep, :] = cell._forward(encoder_hidden_state, decoder_hidden_state[:, timestep, :])
            self.attention_cells.append(cell)
            self.attention_weights = np.append(self.attention_weights, cell.attention)

        return result

    def _backward(self, dout):
        batch_size, n_timestep, hidden_size = dout.shape
        encoder_dhidden_state = 0
        decoder_dhidden_state = np.empty_like(dout).astype(np.float32)

        for timestep in range(n_timestep):
            cell = self.attention_cells[timestep]
            dhidden_state, dh_t = cell._backward(dout[:, timestep, :])
            encoder_dhidden_state += dhidden_state
            decoder_dhidden_state[:, timestep, :] = dh_t

        return encoder_dhidden_state, decoder_dhidden_state


class AttentionEncoder(Model):
    def __init__(self, vocab_size=1000, word_vec_size=128, hidden_size=128, n_layers=1, dropout=None, *layers):
        super().__init__(*layers)
        if len(layers) == 0:
            self.add_layer(EmbeddingLayer(vocab_size=vocab_size, hidden_size=word_vec_size))
            for i in range(n_layers):
                self.add_layer(LSTMLayer(word_vec_size, hidden_size, stateful=False))
                if dropout != None:
                    self.add_layer(Dropout(dropout_ratio=dropout))

    def _forward(self, x):
        hidden_state = self.forward(x)

        return hidden_state

    def _backward(self, dhidden_state):
        result = dhidden_state
        for layer_name in reversed(self.sequence):
            layer = self.network[layer_name]
            result = layer._backward(result)

        return result


class AttentionDecoder(Model):
    def __init__(self, vocab_size=1000, word_vec_size=128, hidden_size=128, n_layers=1, dropout=None, *layers):
        super().__init__(*layers)
        if len(layers) == 0:
            self.add_layer(EmbeddingLayer(vocab_size=vocab_size, hidden_size=word_vec_size, initialize=100))
            for i in range(n_layers):
                self.add_layer(LSTMLayer(word_vec_size, hidden_size))
                if dropout != None:
                    self.add_layer(Dropout(dropout_ratio=dropout))

            self.add_layer(AttentionLayer())
            self.add_layer(DenseLayer(hidden_size * 2, hidden_size))
            self.add_layer(DenseLayer(hidden_size, vocab_size))

    def _forward(self, x, encoder_hidden_state):
        if self.tying_weight:
            self.tying_forward()

        h = encoder_hidden_state[:, -1]
        layer = self.network["LSTMLayer1"]
        layer.load_state(h)

        if self.tying_weight:
            self.tying_forward()

        input = x

        # LSTM까지만 진행
        for layer_name in self.sequence[:-3]:
            layer = self.network[layer_name]
            y = layer._forward(input)
            input = y

        decoder_hidden_state = input

        attention = self.network["AttentionLayer1"]
        context = attention._forward(encoder_hidden_state, decoder_hidden_state)

        concat_dense = self.network["DenseLayer1"]
        final_dense = self.network["DenseLayer2"]

        result = np.concatenate((context, decoder_hidden_state), axis=2).astype(np.float32)
        result = concat_dense._forward(result)

        result = final_dense._forward(result)

        return result

    def _backward(self, dout):
        # dense layer
        concat_dense = self.network["DenseLayer1"]
        final_dense = self.network["DenseLayer2"]
        dout = final_dense._backward(dout)
        dout = concat_dense._backward(dout)

        batch_size, n_timestep, hidden_size_double = dout.shape
        hidden_size = hidden_size_double // 2

        # attention layer
        attention = self.network["AttentionLayer1"]

        dcontext, ddecoder_hidden_state_0 = dout[:, :, :hidden_size], dout[:, :, hidden_size:]

        dencoder_hidden_state, ddecoder_hidden_state_1 = attention._backward(dcontext)
        ddecoder_hidden_state = ddecoder_hidden_state_0 + ddecoder_hidden_state_1

        # lstm부터 끝(embedding)까지
        dout = ddecoder_hidden_state
        for layer_name in reversed(self.sequence[:-3]):
            layer = self.network[layer_name]
            dout = layer._backward(dout)

            # 이후 encoder로 전달받은 gradient 합산해서 전달
        first_lstm_dh = self.network["LSTMLayer1"].dh
        dencoder_hidden_state[:, -1] += first_lstm_dh

        if self.tying_weight:
            self.tying_backward()

        return dencoder_hidden_state

    def generate(self, encoder_hidden_state, start_id, sample_size):
        sampled_result = []
        sample_id = start_id
        h = encoder_hidden_state[:, -1]
        layer = self.network["LSTMLayer1"]
        layer.load_state(deepcopy(h))

        if self.tying_weight:
            self.tying_forward()

        for _ in range(sample_size):
            input = np.array([sample_id]).reshape((1, 1))
            # LSTM까지만 진행
            for layer_name in self.sequence[:-3]:
                layer = self.network[layer_name]
                y = layer._forward(input)
                input = y

            decoder_hidden_state = input

            attention = self.network["AttentionLayer1"]
            context = attention._forward(encoder_hidden_state, decoder_hidden_state)

            concat_dense = self.network["DenseLayer1"]
            final_dense = self.network["DenseLayer2"]

            result = np.concatenate((context, decoder_hidden_state), axis=2).astype(np.float32)
            result = concat_dense._forward(result)

            result = final_dense._forward(result)

            sample_id = np.argmax(result.flatten())
            sampled_result.append(sample_id)

        return sampled_result


class AttentionSeq2Seq(Model):
    def __init__(self, vocab_size=1000, word_vec_size=128, hidden_size=128, n_layers=1, concat=True, dropout=None):
        super().__init__(())
        self.encoder = AttentionEncoder(vocab_size=vocab_size,
                                        word_vec_size=word_vec_size,
                                        hidden_size=hidden_size,
                                        n_layers=n_layers,
                                        dropout=dropout)

        if concat:
            self.decoder = AttentionDecoder(vocab_size=vocab_size,
                                            word_vec_size=word_vec_size,
                                            hidden_size=hidden_size,
                                            n_layers=n_layers,
                                            dropout=dropout)
        else:
            self.decoder = AttentionDecoderBackup(vocab_size=vocab_size,
                                                  word_vec_size=word_vec_size,
                                                  hidden_size=hidden_size,
                                                  n_layers=n_layers,
                                                  dropout=dropout)

        self.network = OrderedDict()
        self.count_dict = OrderedDict()
        self.layers = self.encoder.layers + self.decoder.layers
        self.sequence = []
        temp_repr_list = []

        for layer in self.layers:
            temp_repr_list.append(repr(layer))
        repr_set = set(temp_repr_list)

        # initialize count_dict
        for rep in repr_set:
            self.count_dict[rep] = 1

        for layer in self.layers:
            self.network[f"{repr(layer)}{self.count_dict[repr(layer)]}"] = layer
            self.sequence.append(f"{repr(layer)}{self.count_dict[repr(layer)]}")
            self.count_dict[repr(layer)] += 1

    def __repr__(self) -> str:
        return "Seq2Seq"

    def __call__(self, *args):
        result = self.forward(*args)
        return result

    def forward(self, x, decoder_x):
        h = self.encoder._forward(x)
        result = self.decoder._forward(decoder_x, h)

        return result

    def backward(self, loss):
        dout = loss._backward()
        dh = self.decoder._backward(dout)
        self.encoder._backward(dh)

    def generate(self, x, start_id, sample_size):
        h = self.encoder._forward(x)
        sampled = self.decoder.generate(h,
                                        start_id=start_id,
                                        sample_size=sample_size)

        return sampled

    def weight_tying(self):
        self.decoder.weight_tying()


class AttentionDecoderBackup(Model):
    def __init__(self, vocab_size=1000, word_vec_size=128, hidden_size=128, n_layers=1, dropout=None, *layers):
        super().__init__(*layers)
        if len(layers) == 0:
            self.add_layer(EmbeddingLayer(vocab_size=vocab_size, hidden_size=word_vec_size, initialize=100))
            for i in range(n_layers):
                self.add_layer(LSTMLayer(word_vec_size, hidden_size))
                if dropout != None:
                    self.add_layer(Dropout(dropout_ratio=dropout))

            self.add_layer(AttentionLayer())
            self.add_layer(DenseLayer(hidden_size * 2, vocab_size))

    def _forward(self, x, encoder_hidden_state):
        if self.tying_weight:
            self.tying_forward()

        h = encoder_hidden_state[:, -1]
        layer = self.network["LSTMLayer1"]
        layer.load_state(deepcopy(h))

        if self.tying_weight:
            self.tying_forward()

        input = x

        # LSTM까지만 진행
        for layer_name in self.sequence[:-2]:
            layer = self.network[layer_name]
            y = layer._forward(input)
            input = y

        decoder_hidden_state = input

        attention = self.network["AttentionLayer1"]
        context = attention._forward(encoder_hidden_state, decoder_hidden_state)

        final_dense = self.network["DenseLayer1"]

        if self.tying_weight:
            result = (context + decoder_hidden_state) / 2

        result = np.concatenate((context, decoder_hidden_state), axis=2).astype(np.float32)
        result = final_dense._forward(result)

        return result

    def _backward(self, dout):
        # dense layer
        final_dense = self.network["DenseLayer1"]
        dout = final_dense._backward(dout)

        if self.tying_weight:
            batch_size, n_timestep, hidden_size = dout.shape

        else:
            batch_size, n_timestep, hidden_size_double = dout.shape
            hidden_size = hidden_size_double // 2

        # attention layer
        attention = self.network["AttentionLayer1"]

        if self.tying_weight:
            dcontext = dout / 2
            ddecoder_hidden_state_0 = dout / 2
        else:
            dcontext, ddecoder_hidden_state_0 = dout[:, :, :hidden_size], dout[:, :, hidden_size:]

        dencoder_hidden_state, ddecoder_hidden_state_1 = attention._backward(dcontext)
        ddecoder_hidden_state = ddecoder_hidden_state_0 + ddecoder_hidden_state_1

        # lstm부터 끝(embedding)까지
        dout = ddecoder_hidden_state
        for layer_name in reversed(self.sequence[:-2]):
            layer = self.network[layer_name]
            dout = layer._backward(dout)

            # 이후 encoder로 전달받은 gradient 합산해서 전달
        first_lstm_dh = self.network["LSTMLayer1"].dh
        dencoder_hidden_state[:, -1] += first_lstm_dh

        if self.tying_weight:
            self.tying_backward()

        return dencoder_hidden_state

    def generate(self, encoder_hidden_state, start_id, sample_size):
        sampled_result = []
        sample_id = start_id
        h = encoder_hidden_state[:, -1]
        layer = self.network["LSTMLayer1"]
        layer.load_state(deepcopy(h))

        if self.tying_weight:
            self.tying_forward()

        for _ in range(sample_size):
            input = np.array([sample_id]).reshape((1, 1))
            # LSTM까지만 진행
            for layer_name in self.sequence[:-2]:
                layer = self.network[layer_name]
                y = layer._forward(input)
                input = y

            decoder_hidden_state = input

            attention = self.network["AttentionLayer1"]
            context = attention._forward(encoder_hidden_state, decoder_hidden_state)

            final_dense = self.network["DenseLayer1"]

            result = np.concatenate((context, decoder_hidden_state), axis=2).astype(np.float32)
            result = final_dense._forward(result)

            sample_id = np.argmax(result.flatten())
            sampled_result.append(sample_id)

        return sampled_result


import pickle
from copy import deepcopy
from collections import OrderedDict
from function import *
from function_class import *
from gpu import *
from utils import *
from model import *


class BaseModule():
    def __init__(self):
        self.parameter = OrderedDict()
        self.differentiable = False
        self.changeability = False
        self.mixed_precision = False
        self.tying = False
        self.tied = False

    def __repr__(self):
        return "Module"

    def _save_params(self):
        """Layer의 parameter 반환

        Returns:
            OrderedDict
        """
        return deepcopy(self.parameter)

    def _load_params(self, params):
        self.parameter = deepcopy(params)

    def _to_cpu(self):
        """Layer의 parameter를 vram에서 ram으로 옮김
        """
        param_list = list(self.parameter.keys())
        for param in param_list:
            self.parameter[param] = to_cpu(self.parameter[param])

    def _to_gpu(self):
        """Layer의 parameter를 ram에서 vram으로 옮김
        """
        param_list = list(self.parameter.keys())
        for param in param_list:
            self.parameter[param] = to_gpu(self.parameter[param])


class ResNet34Module(BaseModule):
    def __init__(self, input_channel, output_channel, kernel_size, stride, padding, residual_connect=True,
                 initialize: str = "He"):
        super().__init__()
        self.conv1 = ConvLayer(input_channel, output_channel, kernel_size, stride, padding)
        self.conv2 = ConvLayer(output_channel, output_channel, kernel_size, stride, padding)
        self.batch_norm1 = BatchNorm2D(output_channel)
        self.batch_norm2 = BatchNorm2D(output_channel)
        self.relu1 = ReLU()
        self.relu2 = ReLU()
        self.residual_connect = residual_connect
        self.first_forwarded = False

        self.differentiable = True
        self.changeability = True
        self.bn_not_initialized = True

        if isinstance(kernel_size, int):
            self.kernel_width = kernel_size
            self.kernel_height = kernel_size

        elif isinstance(kernel_size, tuple):
            self.kernel_height, self.kernel_width = kernel_size

        self.conv1_fan_in = self.kernel_width * self.kernel_height * input_channel
        self.conv1_fan_out = self.kernel_width * self.kernel_height * output_channel
        self.conv2_fan_in = self.kernel_width * self.kernel_height * output_channel
        self.conv2_fan_out = self.kernel_width * self.kernel_height * output_channel

        if initialize == "He":
            self.parameter["conv1_weight"] = np.random.randn(output_channel, input_channel, self.kernel_height,
                                                             self.kernel_width).astype(np.float32) * (
                                                 np.sqrt(2 / self.conv1_fan_in))
            self.parameter["conv1_bias"] = np.zeros(output_channel).astype(np.float32)
            self.parameter["conv2_weight"] = np.random.randn(output_channel, output_channel, self.kernel_height,
                                                             self.kernel_width).astype(np.float32) * (
                                                 np.sqrt(2 / self.conv2_fan_in))
            self.parameter["conv2_bias"] = np.zeros(output_channel).astype(np.float32)

        elif initialize == "Xavier":
            self.parameter["conv1_weight"] = np.random.randn(output_channel, input_channel, self.kernel_height,
                                                             self.kernel_width).astype(np.float32) * (
                                                 np.sqrt(1 / self.conv1_fan_in))
            self.parameter["conv1_bias"] = np.zeros(output_channel).astype(np.float32)
            self.parameter["conv2_weight"] = np.random.randn(output_channel, output_channel, self.kernel_height,
                                                             self.kernel_width).astype(np.float32) * (
                                                 np.sqrt(1 / self.conv2_fan_in))
            self.parameter["conv2_bias"] = np.zeros(output_channel).astype(np.float32)

        elif initialize == "None":
            self.parameter["conv1_weight"] = 0.01 * np.random.randn(output_channel, input_channel, self.kernel_height,
                                                                    self.kernel_width).astype(np.float32)
            self.parameter["conv1_bias"] = np.zeros(output_channel).astype(np.float32)
            self.parameter["conv2_weight"] = 0.01 * np.random.randn(output_channel, output_channel, self.kernel_height,
                                                                    self.kernel_width).astype(np.float32)
            self.parameter["conv2_bias"] = np.zeros(output_channel).astype(np.float32)

        else:
            raise ValueError("'initialize' must be 'He' or 'Xavier' or 'None'")

        self.conv1.parameter["weight"] = self.parameter["conv1_weight"]
        self.conv1.parameter["bias"] = self.parameter["conv1_bias"]
        self.conv2.parameter["weight"] = self.parameter["conv2_weight"]
        self.conv2.parameter["bias"] = self.parameter["conv2_bias"]
        self.parameter["b1_gamma"]
        self.parameter["b1_beta"]
        self.parameter["b2_gamma"]
        self.parameter["b2_beta"]

    def sync_param(self):
        self.parameter["conv1_weight"] = self.conv1.parameter["weight"]
        self.parameter["conv1_bias"] = self.conv1.parameter["bias"]
        self.parameter["conv2_weight"] = self.conv2.parameter["weight"]
        self.parameter["conv2_bias"] = self.conv2.parameter["bias"]
        self.parameter["b1_gamma"] = self.batch_norm1.parameter["gamma"]
        self.parameter["b1_beta"] = self.batch_norm1.parameter["beta"]
        self.parameter["b2_gamma"] = self.batch_norm2.parameter["gamma"]
        self.parameter["b2_beta"] = self.batch_norm2.parameter["beta"]

    def __repr__(self):
        return "ResidualModule"

    def __call__(self, arg):
        result = self._forward(arg)
        return result

    def _forward(self, x):
        result = self.conv1(x)
        result = self.batch_norm1(result)
        result = self.relu1(result)
        result = self.conv2(result)
        result = self.batch_norm2(result)
        if self.residual_connect:
            result = np.add(self.relu2(result), x)
        else:
            result = self.relu2(result)

        if self.bn_not_initialized:
            self.parameter["b1_gamma"] = self.batch_norm1.parameter["gamma"]
            self.parameter["b1_beta"] = self.batch_norm1.parameter["beta"]

            self.parameter["b2_gamma"] = self.batch_norm2.parameter["gamma"]
            self.parameter["b2_beta"] = self.batch_norm2.parameter["beta"]
            self.bn_not_initialized = False

        return result

    def _backward(self, dout):

        result = self.relu2._backward(dout)
        result = self.batch_norm2._backward(result)
        result = self.conv2._backward(result)
        result = self.relu1._backward(result)
        result = self.batch_norm1._backward(result)
        result = self.conv1._backward(result)

        return result

    def get_gradient(self) -> OrderedDict:
        """Layer의 gradient를 return

        Returns:
            OrderedDict: Layer의 gradient
        """
        grad = OrderedDict()
        grad["c1_dw"] = self.conv1.dw
        grad["c1_db"] = self.conv1.db
        grad["c2_dw"] = self.conv2.dw
        grad["c2_db"] = self.conv2.db
        grad["b1_dgamma"] = self.batch_norm1.dgamma
        grad["b1_dbeta"] = self.batch_norm1.dbeta
        grad["b2_dgamma"] = self.batch_norm2.dgamma
        grad["b2_dbeta"] = self.batch_norm2.dbeta

        return grad

    def eval_state(self):
        self.batch_norm1.eval_state()
        self.batch_norm2.eval_state()

    def train_state(self):
        self.batch_norm1.train_state()
        self.batch_norm2.train_state()

    def sync_param(self):
        self.parameter["conv1_weight"] = self.conv1.parameter["weight"]
        self.parameter["conv1_bias"] = self.conv1.parameter["bias"]
        self.parameter["conv2_weight"] = self.conv2.parameter["weight"]
        self.parameter["conv2_bias"] = self.conv2.parameter["bias"]
        self.parameter["b1_gamma"] = self.batch_norm1.parameter["gamma"]
        self.parameter["b1_beta"] = self.batch_norm1.parameter["beta"]
        self.parameter["b2_gamma"] = self.batch_norm2.parameter["gamma"]
        self.parameter["b2_beta"] = self.batch_norm2.parameter["beta"]


class ResNet34(Model):
    def __init__(self, input_size, num_class):
        super().__init__()
        if isinstance(input_size, int):
            self.input_width = input_size
            self.input_height = input_size

        elif isinstance(input_size, tuple):
            self.input_height, self.kernel_width = input_size

        self.add_layer(ConvLayer(3, 64, (7, 7), stride=2, padding=3),
                       BatchNorm2D(64),
                       ReLU(),
                       MaxPoolingLayer(3, stride=2, padding=1)
                       )

        self.add_layer(
            ResNet34Module(64, 64, kernel_size=(3, 3), stride=1, padding=1),
            ResNet34Module(64, 64, kernel_size=(3, 3), stride=1, padding=1),
            ResNet34Module(64, 64, kernel_size=(3, 3), stride=1, padding=1)
        )

        self.add_layer(
            ResNet34Module(64, 128, kernel_size=(3, 3), stride=1, padding=1, residual_connect=False),
            ResNet34Module(128, 128, kernel_size=(3, 3), stride=1, padding=1),
            ResNet34Module(128, 128, kernel_size=(3, 3), stride=1, padding=1),
            ResNet34Module(128, 128, kernel_size=(3, 3), stride=1, padding=1)
        )

        self.add_layer(
            ResNet34Module(128, 256, kernel_size=(3, 3), stride=1, padding=1, residual_connect=False),
            ResNet34Module(256, 256, kernel_size=(3, 3), stride=1, padding=1),
            ResNet34Module(256, 256, kernel_size=(3, 3), stride=1, padding=1),
            ResNet34Module(256, 256, kernel_size=(3, 3), stride=1, padding=1),
            ResNet34Module(256, 256, kernel_size=(3, 3), stride=1, padding=1),
            ResNet34Module(256, 256, kernel_size=(3, 3), stride=1, padding=1)
        )

        self.add_layer(
            ResNet34Module(256, 512, kernel_size=(3, 3), stride=1, padding=1, residual_connect=False),
            ResNet34Module(512, 512, kernel_size=(3, 3), stride=1, padding=1),
            ResNet34Module(512, 512, kernel_size=(3, 3), stride=1, padding=1)
        )

        self.add_layer(
            GlobalAveragePoolingLayer(),
            DenseLayer(512, 1000),
            BatchNorm1D(1000),
            ReLU(),
            Dropout(0.3),
            DenseLayer(1000, num_class)
        )


class VGG16(Model):
    def __init__(self, input_size, num_class, input_channel=3):
        super().__init__()
        if isinstance(input_size, int):
            self.input_width = input_size
            self.input_height = input_size

        elif isinstance(input_size, tuple):
            self.input_height, self.kernel_width = input_size

        self.add_layer(ConvLayer(input_channel, 64, (3, 3), stride=1, padding=1),
                       BatchNorm2D(64),
                       ReLU(),
                       ConvLayer(64, 64, (3, 3), stride=1, padding=1),
                       BatchNorm2D(64),
                       ReLU(),
                       MaxPoolingLayer((2, 2), stride=2))

        self.add_layer(ConvLayer(64, 128, (3, 3), stride=1, padding=1),
                       BatchNorm2D(128),
                       ReLU(),
                       ConvLayer(128, 128, (3, 3), stride=1, padding=1),
                       BatchNorm2D(128),
                       ReLU(),
                       MaxPoolingLayer((2, 2), stride=2))

        self.add_layer(ConvLayer(128, 256, (3, 3), stride=1, padding=1),
                       BatchNorm2D(256),
                       ReLU(),
                       ConvLayer(256, 256, (3, 3), stride=1, padding=1),
                       BatchNorm2D(256),
                       ReLU(),
                       ConvLayer(256, 256, (3, 3), stride=1, padding=1),
                       BatchNorm2D(256),
                       ReLU(),
                       MaxPoolingLayer((2, 2), stride=2))

        self.add_layer(ConvLayer(256, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       MaxPoolingLayer((2, 2), stride=2))

        self.add_layer(ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       MaxPoolingLayer((2, 2), stride=2))

        self.add_layer(DenseLayer(512 * (int(self.input_height // 32) * int(self.input_height // 32)), 64),
                       ReLU(),
                       Dropout(0.5),
                       DenseLayer(64, 64),
                       ReLU(),
                       Dropout(0.5),
                       DenseLayer(64, num_class))


class VGG19(Model):
    def __init__(self, input_size, num_class, input_channel=3):
        super().__init__()
        if isinstance(input_size, int):
            self.input_width = input_size
            self.input_height = input_size

        elif isinstance(input_size, tuple):
            self.input_height, self.kernel_width = input_size

        self.add_layer(
            ConvLayer(input_channel, 64, (3, 3), stride=1, padding=1),
            BatchNorm2D(64),
            ReLU(),
            ConvLayer(64, 64, (3, 3), stride=1, padding=1),
            BatchNorm2D(64),
            ReLU(),
            MaxPoolingLayer((2, 2), stride=2)
        )

        self.add_layer(ConvLayer(64, 128, (3, 3), stride=1, padding=1),
                       BatchNorm2D(128),
                       ReLU(),
                       ConvLayer(128, 128, (3, 3), stride=1, padding=1),
                       BatchNorm2D(128),
                       ReLU(),
                       MaxPoolingLayer((2, 2), stride=2))

        self.add_layer(ConvLayer(128, 256, (3, 3), stride=1, padding=1),
                       BatchNorm2D(256),
                       ReLU(),
                       ConvLayer(256, 256, (3, 3), stride=1, padding=1),
                       BatchNorm2D(256),
                       ReLU(),
                       ConvLayer(256, 256, (3, 3), stride=1, padding=1),
                       BatchNorm2D(256),
                       ReLU(),
                       ConvLayer(256, 256, (3, 3), stride=1, padding=1),
                       BatchNorm2D(256),
                       ReLU(),
                       MaxPoolingLayer((2, 2), stride=2))

        self.add_layer(ConvLayer(256, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       MaxPoolingLayer((2, 2), stride=2))

        self.add_layer(ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       ConvLayer(512, 512, (3, 3), stride=1, padding=1),
                       BatchNorm2D(512),
                       ReLU(),
                       MaxPoolingLayer((2, 2), stride=2))

        self.add_layer(DenseLayer(512 * (int(self.input_height // 32) * int(self.input_height // 32)), 64),
                       ReLU(),
                       Dropout(0.5),
                       DenseLayer(64, 64),
                       ReLU(),
                       Dropout(0.5),
                       DenseLayer(64, num_class))


class LeNet(Model):
    def __init__(self, input_size, num_class, input_channel=3):
        super().__init__()
        if isinstance(input_size, int):
            self.input_width = input_size
            self.input_height = input_size

        elif isinstance(input_size, tuple):
            self.input_height, self.kernel_width = input_size

        self.add_layer(
            ConvLayer(input_channel, 6, (5, 5), stride=1, padding=0),
            BatchNorm2D(6),
            ReLU(),
            MaxPoolingLayer((2, 2), stride=2)
        )

        self.add_layer(
            ConvLayer(6, 16, (5, 5), stride=1, padding=0),
            BatchNorm2D(16),
            ReLU(),
            MaxPoolingLayer((2, 2), stride=2))

        self.add_layer(
            ConvLayer(16, 120, (5, 5), stride=1, padding=0),
            BatchNorm2D(120),
            ReLU()
        )

        self.add_layer(
            DenseLayer(120, 84),
            ReLU(),
            Dropout(0.3),
            DenseLayer(84, num_class)
        )





from gpu import *


class DataLoader:
    def __init__(self, x, y, batch_size=1, random_sampling=False):
        """
        Build dataloader that can iterate with batch size in dataset (for supervised task)

        Parameters
        ----------
        x : data corresponding to feature

        y : data corresponding to label

        batch_size (int) : batch size. Default: 1

        random_sampling (bool) : Whether for random sampling from dataset. Default: False

        """
        self.current = 0
        self.x = np.array(x)
        self.y = np.array(y)
        self.batch_size = batch_size
        self.batch_mask = np.arange(self.x.shape[0])
        if random_sampling:
            np.random.shuffle(self.batch_mask)

    def __iter__(self):
        return self

    def __next__(self):
        if self.current + self.batch_size <= len(self.x):
            x = self.x[self.batch_mask[self.current:self.current + self.batch_size]]
            y = self.y[self.batch_mask[self.current:self.current + self.batch_size]]
            self.current += self.batch_size
            return x, y

        elif self.current >= len(self.batch_mask):
            # for next epoch
            self.current = 0
            raise StopIteration

        else:
            x = self.x[self.batch_mask[self.current:]]
            y = self.y[self.batch_mask[self.current:]]
            self.current += self.batch_size
            return x, y

    def __getitem__(self, index):
        return self.x[index], self.y[index]

    def __len__(self):
        return int(len(self.x) / self.batch_size)

    def dataset_len(self):
        return len(self.x)


class LanguageModelingDataLoader():
    def __init__(self, corpus, batch_size=1, time_size=10):
        """
        Build dataloader that can iterate with batch size in dataset (for Language Modeling)

        Parameters
        ----------
        corpus : data corresponding to feature

        batch_size (int) : batch size. Default: 1

        """
        self.current_iter = 0
        self.corpus = np.array(corpus)
        self.x_temp = np.array(corpus[:-1])
        self.y_temp = np.array(corpus[1:])
        self.corpus_size = len(corpus)
        self.data_size = len(self.x_temp)
        self.batch_size = batch_size
        self.time_size = time_size
        self.jump = (self.corpus_size - 1) // self.batch_size
        # offset : 각각의 mini-batch의 시작 요소의 위치를 담음
        self.offset = [i * self.jump for i in range(self.batch_size)]
        self.max_iteration = self.corpus_size // (self.batch_size * self.time_size)
        self.time_index = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.current_iter < self.max_iteration - 1:
            x = np.empty((self.batch_size, self.time_size), dtype="i")
            y = np.empty((self.batch_size, self.time_size), dtype="i")
            for t in range(self.time_size):
                for i, offset in enumerate(self.offset):
                    x[i, t] = self.x_temp[(offset + self.time_index) % self.data_size]
                    y[i, t] = self.y_temp[(offset + self.time_index) % self.data_size]
                self.time_index += 1
            self.current_iter += 1
            return x, y

        elif self.current_iter >= self.max_iteration - 1:
            # for next epoch
            self.current_iter = 0
            raise StopIteration

    def __getitem__(self, index):
        return self.x_temp[index], self.y_temp[index]

    def __len__(self):
        return self.max_iteration

    def dataset_len(self):
        return self.corpus_size



from collections import Counter
from gpu import *
from function import *
from model import *

class BaseFunction():
    def __init__(self):
        pass

    def __call__(self, arg):
        result = self._forward(arg)
        return result

    def _forward(self, x):
        pass

    def __repr__(self) -> str:
        return "Function"


class Step(BaseFunction):
    def __init__(self):
        super().__init__()

    def _forward(self, x):
        return step_function(x)


class Identity(BaseFunction):
    def __init__(self):
        super().__init__()

    def _forward(self, x):
        return identity(x)


class Sigmoid(BaseFunction):
    def __init__(self):
        self.differentiable = False
        self.out = None
        super().__init__()

    def _forward(self, x):
        out = sigmoid(x)
        self.out = out

        return out

    def _backward(self, input):
        result = input * self.out * (1.0 - self.out)

        return result


class Softmax(BaseFunction):
    def __init__(self):
        super().__init__()

    def forward(self, x):
        return softmax(x)


class ReLU(BaseFunction):
    def __init__(self):
        self.differentiable = False
        self.mask = None
        super().__init__()

    def _forward(self, x):
        self.mask = (x <= 0)
        result = x.copy()
        result[self.mask] = 0

        return result

    def _backward(self, input):
        input[self.mask] = 0
        result = input

        return result


class CrossEntropyLoss():
    def __init__(self):
        self.loss = None
        self.pred = None
        self.true = None
        self.cache = None
        self.ignore_label = -1

    def __repr__(self):
        return "Function"

    def __call__(self, *arg):
        result = self._forward(*arg)
        return result

    def _forward(self, pred, true):
        if pred.ndim == 3:
            batch_size, n_timestep, vocab_size = pred.shape

            if true.ndim == 3:
                true = true.argmax(axis=2)

            mask = (true != self.ignore_label)

            self.pred = pred.reshape(batch_size * n_timestep, vocab_size)
            self.true = true.reshape(batch_size * n_timestep)
            mask = mask.reshape(batch_size * n_timestep)

            pred_sentence = softmax(self.pred)
            ls = np.log(pred_sentence[np.arange(batch_size * n_timestep), self.true])
            ls *= mask
            self.loss = -np.sum(ls)
            self.loss /= mask.sum()

            self.cache = (self.true, pred_sentence, mask, (batch_size, n_timestep, vocab_size))

            return self.loss
        else:
            self.pred = softmax(pred)
            self.true = true
            self.loss = cross_entropy_error(self.pred, self.true)

            return self.loss

    def _backward(self):
        if self.cache is not None:
            true, pred_sentence, mask, (batch_size, n_timestep, vocab_size) = self.cache

            dx = pred_sentence
            dx[np.arange(batch_size * n_timestep), true] -= 1
            dx *= 1
            dx /= mask.sum()
            dx *= mask[:, np.newaxis]  # ignore_label에 해당하는 데이터는 기울기를 0으로 설정

            dx = dx.reshape((batch_size, n_timestep, vocab_size))

            return dx

        batch_size = self.true.shape[0]
        if self.true.size == self.pred.size:
            dx = (self.pred - self.true) / batch_size
        else:
            dx = self.pred.copy()
            dx[np.arange(batch_size), self.true] -= 1
            dx = dx / batch_size

        return dx


class BinaryCrossEntropyLoss():
    def __init__(self):
        self.loss = None
        self.pred = None
        self.true = None

    def __repr__(self) -> str:
        return "Function"

    def __call__(self, *arg):
        result = self._forward(*arg)
        return result

    def _forward(self, pred, true):
        self.pred = sigmoid(pred)
        self.true = true
        self.loss = cross_entropy_error(self.pred, self.true)

        return self.loss

    def _backward(self):
        batch_size = self.true.shape[0]
        if self.true.size == self.pred.size:
            dx = (self.pred - self.true) / batch_size
        else:
            dx = self.pred.copy()
            dx[np.arange(batch_size), self.true] -= 1
            dx = dx / batch_size

        return dx



from gpu import *
from collections import OrderedDict


class BaseOptimizer():
    def __init__(self, lr=0.01):
        self.lr = lr
        self.param_grad_dict = OrderedDict()
        self.param_grad_dict["weight"] = "dw"
        self.param_grad_dict["weight_x"] = "dwx"
        self.param_grad_dict["weight_h"] = "dwh"
        self.param_grad_dict["bias"] = "db"
        self.param_grad_dict["gamma"] = "dgamma"
        self.param_grad_dict["beta"] = "dbeta"
        self.param_grad_dict["conv1_weight"] = "c1_dw"
        self.param_grad_dict["conv1_bias"] = "c1_db"
        self.param_grad_dict["conv2_weight"] = "c2_dw"
        self.param_grad_dict["conv2_bias"] = "c2_db"
        self.param_grad_dict["b1_gamma"] = "b1_dgamma"
        self.param_grad_dict["b1_beta"] = "b1_dbeta"
        self.param_grad_dict["b2_gamma"] = "b2_dgamma"
        self.param_grad_dict["b2_beta"] = "b2_dbeta"

    def __repr__(self):
        return "Optimizer"


class SGDOptimizer(BaseOptimizer):
    def __init__(self, lr=0.01):
        super().__init__(lr)

    def update(self, model):
        # update parameter
        for layer_name in model.sequence:
            layer = model.network[layer_name]

            # only update differentiable layer
            # do not update tied layer
            if layer.differentiable == True and layer.tied == False:
                grad = layer.get_gradient()
                param_list = list(layer.parameter.keys())

                for param in param_list:
                    model.network[layer_name].parameter[param] -= (self.lr * grad[self.param_grad_dict[param]])


class MomentumOptimizer(BaseOptimizer):
    def __init__(self, lr=0.01, momentum=0.9):
        super().__init__(lr)
        self.momentum = momentum
        self.v = None

    def update(self, model):
        if self.v is None:

            # initialize v
            self.v = OrderedDict()
            for layer_name in model.sequence:
                layer = model.network[layer_name]
                if layer.differentiable == True and layer.tied == False:
                    param_list = list(layer.parameter.keys())
                    self.v[layer_name] = OrderedDict()

                    for param in param_list:
                        self.v[layer_name][param] = np.zeros_like(layer.parameter[param])

        # update parameter
        for layer_name in model.sequence:
            layer = model.network[layer_name]

            # only update differentiable layer
            # do not update tied layer
            if layer.differentiable == True and layer.tied == False:
                param_list = list(layer.parameter.keys())
                grad = layer.get_gradient()

                for param in param_list:
                    self.v[layer_name][param] = self.momentum * self.v[layer_name][param] - (
                                self.lr * grad[self.param_grad_dict[param]])
                    model.network[layer_name].parameter[param] += self.v[layer_name][param]


class AdaGrad(BaseOptimizer):
    def __init__(self, lr=0.01):
        super().__init__(lr)
        self.h = None

    def update(self, model):
        if self.h is None:
            # initialize h
            self.h = OrderedDict()
            for layer_name in model.sequence:
                layer = model.network[layer_name]
                if layer.differentiable == True and layer.tied == False:
                    self.h[layer_name] = OrderedDict()
                    param_list = list(layer.parameter.keys())

                    for param in param_list:
                        self.h[layer_name][param] = np.zeros_like(layer.parameter[param])

        # update parameter
        for layer_name in model.sequence:
            layer = model.network[layer_name]

            # only update differentiable layer
            # do not update tied layer
            if layer.differentiable == True and layer.tied == False:
                grad = layer.get_gradient()
                param_list = list(layer.parameter.keys())

                for param in param_list:
                    # update h
                    self.h[layer_name][param] = grad[self.param_grad_dict[param]] * grad[self.param_grad_dict[param]]
                    # update parameter
                    model.network[layer_name].parameter[param] -= self.lr * grad[self.param_grad_dict[param]] / (
                                np.sqrt(self.h[layer_name][param]) + 1e-7)


class Adam(BaseOptimizer):
    def __init__(self, lr=0.001, b1=0.9, b2=0.999, epsilon=1e-8, loss_scaling=None):
        super().__init__(lr)
        self.b1 = b1
        self.b2 = b2
        self.epsilon = epsilon
        self.m = None
        self.v = None
        self.iter = 0
        self.loss_scaling = loss_scaling

    def update(self, model):
        if self.m is None:

            # initialize m, v
            self.m, self.v = OrderedDict(), OrderedDict()
            for layer_name in model.sequence:
                layer = model.network[layer_name]
                if layer.differentiable == True and layer.tied == False:
                    self.m[layer_name], self.v[layer_name] = OrderedDict(), OrderedDict()
                    param_list = list(layer.parameter.keys())

                    for param in param_list:
                        self.m[layer_name][param] = np.zeros_like(layer.parameter[param])
                        self.v[layer_name][param] = np.zeros_like(layer.parameter[param])

        self.iter += 1

        for layer_name in model.sequence:
            layer = model.network[layer_name]

            # only update differentiable layer
            # do not update tied layer
            if layer.differentiable == True and layer.tied == False:

                grad = layer.get_gradient()
                param_list = list(layer.parameter.keys())

                if self.loss_scaling != None:
                    for g in grad.keys():
                        grad[g] *= (1 / self.loss_scaling)

                for param in param_list:
                    # update m, v
                    self.m[layer_name][param] = self.b1 * self.m[layer_name][param] + (1 - self.b1) * grad[
                        self.param_grad_dict[param]]
                    self.v[layer_name][param] = self.b2 * self.v[layer_name][param] + (1 - self.b2) * (
                                grad[self.param_grad_dict[param]] ** 2)

                    # update m_hat, v_hat
                    m_hat = self.m[layer_name][param] / (1 - (self.b1 ** self.iter))
                    v_hat = self.v[layer_name][param] / (1 - (self.b2 ** self.iter))

                    # update parameter
                    model.network[layer_name].parameter[param] -= self.lr / (np.sqrt(v_hat) + self.epsilon) * m_hat



import requests
import json

# 발행한 토큰 불러오기
with open("./token.json", "r") as kakao:
    tokens = json.load(kakao)

url = "https://kapi.kakao.com/v2/api/talk/memo/default/send"

headers = {
    "Authorization": "Bearer " + tokens["access_token"]
}

data = {
    'object_type': 'text',
    'text': '1epoch 종료',
    'link': {
        'web_url': 'https://developers.kakao.com',
        'mobile_web_url': 'https://developers.kakao.com'
    },
    'button_title': 'epoch alert'
}


data = {'template_object': json.dumps(data)}
response = requests.post(url, headers=headers, data=data)
print(response.status_code)
print(response.text)




from gpu import *

def to_cpu(x):
    import numpy
    if type(x) == numpy.ndarray:
        return x
    return np.asnumpy(x)


def to_gpu(x):
    import cupy
    if type(x) == cupy.ndarray:
        return x
    return cupy.asarray(x)


def smooth_curve(x):

    window_len = 11
    s = np.r_[x[window_len-1:0:-1], x, x[-1:-window_len:-1]]
    w = np.kaiser(window_len, 2)
    y = np.convolve(w/w.sum(), s, mode='valid')
    return y[5:len(y)-5]


import requests
import json

url = 'https://kauth.kakao.com/oauth/token'
client_id = ''
redirect_uri = 'https://example.com/oauth'
code = ''

data = {
    'grant_type':'authorization_code',
    'client_id':client_id,
    'redirect_uri':redirect_uri,
    'code': code,
    }

response = requests.post(url, data=data)
tokens = response.json()

#발행된 토큰 저장
with open("token.json","w") as kakao:
    json.dump(tokens, kakao)


import setuptools
from setuptools import setup


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='neuralflow',  # module 이름
    version='0.2.4',
    description='Deep learning framework built with numpy (cupy)',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Sangam Lee',
    author_email='2bhambitious@gmail.com',
    url='https://github.com/augustinLib/neuralflow',
    license='MIT',
    packages=setuptools.find_packages(),
    py_modules=['pymodule'],  # 업로드할 module
    python_requires='>=3.8',  # 파이썬 버전
  # module 필요한 다른 module
    install_requires = [
        "cupy-cuda11x==11.5.0",
        "matplotlib==3.6.3",
        "matplotlib-inline==0.1.6",
        "numpy==1.24.1",
        "packaging==23.0",
        "pandas==1.5.3",
        "pickleshare==0.7.5",
        "Pillow==9.4.0",
        "tqdm==4.64.1"
    ],
    classifiers = [
                      "Programming Language :: Python :: 3",
                      "License :: OSI Approved :: MIT License",
                      "Operating System :: OS Independent"
    ]
)
